"""Classes and methods for working with velocity distributions
within the `dress` framework."""


import numpy as np
from scipy.constants import c
from scipy.interpolate import interp1d

from dress import relkin
from dress import sampler
from dress import vec_ops as vec


class VelocityDistribution:
    """A class representing the velocity distribution of an ensemble of particles.

    In order to use it, one would typically subclass this class and implement
    a method called `_sample(self, n, index=0)`, which returns `n` velocity samples from
    the distribution at the given spatial index.

    Attributes
    ----------
    particle : instance of dress.reactions.Particle
        The particle that the distribution represent

    density : array of shape (N,)
        The density (in particles/m**3) of particles in the distribution,
        at each of the N spatial locations.

    v_collective : array with shape (3,N)
        Collective velocity (m/s) of all particles in the distribution, at each spatial location.
        `v_collective=None` means no collective motion.

    n_spatial : int
        The number of spatial points the distribution is given at. This is typically inferred 
        from the shapes of the distribution data supplied to the constructors of the various 
        sub-classes below.

    pos : tuple of arrays
        Coordinates of each spatial point, e.g. (R,Z) or (X,Y,Z).
        Not necessary for any of the sampling routines, but could be useful
        for the user when plotting etc."""

    def __init__(self, particle, density=None, v_collective=None):
        """Create velocity distribution.

        Parameters
        ----------
        particle : instance of dress.particles.Particle
            The particle that the distribution represent

        density : float or array-like
            The density (in particles/m**3) of particles in the distribution,
            at each of the N spatial locations.

        v_collective : array-like with shape (3,N)
            Collective velocity (m/s) of all particles in the distribution, at each spatial location.
            `v_collective=None` means no collective motion."""
        
        self.particle = particle
        self.density = np.atleast_1d(density)
        self.v_collective = v_collective

        self.pos = None

    def sample(self, n, index=0):
        """Sample `n` velocity vectors from the distribution.

        Parameters
        ----------
        n : int
            The number of samples to draw.
        index : int
            The spatial index of the `dist` array to sample from

        Returns
        -------
        v : array of shape (3,n)
            The sampled velocity vectors (in m/s)."""

        n = int(n)
        v = self._sample(n, index=index)
        
        if self.v_collective is not None:
            v_coll = vec.make_vector(self.v_collective[:,index])
            v = v + v_coll
        
        return v

    def _sample(self, n, index=0):
        """Sampling method, to be overloaded by subclasses."""
        return np.zeros(3,n)

    @property
    def v_collective(self):
        return self._v_collective
    
    @v_collective.setter
    def v_collective(self, v):
        
        if v is None:
            self._v_collective = None
            return

        v = vec.make_vector(v)
        
        if v.shape[1] != len(self.density):
            v = vec.repeat(v, len(self.density))

        self._v_collective = v

    def _set_1d_dist(self, dist):
        """Set the `dist` attribute from the given input array."""

        dist = np.atleast_1d(dist)
        
        if dist.ndim == 1:
            dist = dist[np.newaxis, :]   # dummy spatial axis

        self.dist = dist
        self.n_spatial = len(dist)

    def _set_2d_dist(self, dist):
        """Set the `dist` attribute from the given input array."""

        dist = np.atleast_2d(dist)
        
        if dist.ndim == 2:
            dist = dist[np.newaxis, :, :]   # dummy spatial axis

        self.dist = dist
        self.n_spatial = len(dist)
        

class VparVperpDistribution(VelocityDistribution):
    """A velocity distribution given in terms of the velocity components parallel and
    perpendicular relative to a given reference direction.

    Typically, a user would sub-class this class and override the `sample_vpar_vperp` method.

    Attributes
    ----------
    ref_dir : array with shape (3,N)
        The parallel and perpendicular speeds are given relative to this direction.

    For the rest of the attributes see docstring of the parent class(es)."""

    def __init__(self, particle, density=None, v_collective=None, ref_dir=[0,1,0]):
        
        super().__init__(particle, density=density, v_collective=v_collective)
        self.ref_dir = ref_dir

    @property
    def ref_dir(self):
        return self._ref_dir

    @ref_dir.setter
    def ref_dir(self, u):
        u = vec.make_vector(u)

        if u.shape[1] != len(self.density):
            u = vec.repeat(u, len(self.density))

        self._ref_dir = vec.normalize(u)

        # Construct coordinate system where one axis aligns with the reference direction
        self._set_local_basis_vectors()

    def _sample(self, n, index=0):
        """Sample paralell and perpendicular speeds and calculate corresponding velocities."""
        
        v_par, v_perp = self.sample_vpar_vperp(n, index=index)
        v = self._calc_velocity(v_par, v_perp, index=index)
        
        return v

    def sample_vpar_vperp(self, n, index=0):
        """Sample paralell and perpendicular speeds (m/s). To be overridden by sub-classes."""
        n = int(n)
        return np.zeros(n), np.zeros(n)

    def _set_local_basis_vectors(self):
        """Construct a rectangular coordinate system with basis vectors 
        e1, e2, e3, where the e2 vector is parallel to self.ref_dir."""

        # ref_dir is already normalized to unity
        e2 = self.ref_dir
        
        # There are infinitely many possible choices for e1 and e2 (we can choose any 
        # of them since we will anyway sample the azimuthal angle uniformly). We start 
        # by finding a vector that is non-parallel to e2.
        u = e2 + vec.make_vector([1,0,0])
        u = vec.normalize(u)
        
        # Now e1 and e3 can be constructed by taking appropriate cross products with e2
        e1 = vec.cross(e2, u)
        e1 = vec.normalize(e1)
        e3 = vec.cross(e1, e2)

        self.e1 = e1
        self.e2 = e2
        self.e3 = e3

    def _calc_velocity(self, v_par, v_perp, index=0):
        """Calculate velocity corresponding to given values of paralell 
        and perpendicular speeds (m/s)."""

        n_samples = len(v_par)

        e1 = vec.make_vector(self.e1[:,index])
        e2 = vec.make_vector(self.e2[:,index])
        e3 = vec.make_vector(self.e3[:,index])

        # Parallel velocity is given with respect to the e2 basis vector of the 
        # local coordinate system.
        v_par = v_par*e2

        # Azimuthal angle of perpendicular velocity is taken to be uniformly distributed
        phi = sampler.sample_uniform([0,2*np.pi], n_samples)
        v_perp = v_perp*(np.cos(phi)*e1 - np.sin(phi)*e3)

        return v_par + v_perp


class EnergyPitchDistribution(VparVperpDistribution):
    """A velocity distribution characterized by its distribution in energy and pitch.

    The speed is determined from the energy and the direction of the velocity vector
    is determined from the pitch, defined as

        pitch = v_parallel/v

    where v_parallel is the component of the velocity relative to a given 
    reference direction.

    Typically, a user would sub-class this class and override the 
    `sample_energy_pitch` methods.

    For the rest of the attributes see docstring of the parent class(es)."""

    def __init__(self, particle, density=None, v_collective=None, ref_dir=[0,1,0]):

        super().__init__(particle, density=density, v_collective=v_collective, ref_dir=ref_dir)
        self.ref_dir = ref_dir

    def sample_vpar_vperp(self, n, index=0):
        """Sample (energy,pitch) and evaluate the corresponding (v_par,v_perp), in m/s."""
        E, pitch = self.sample_energy_pitch(n, index=index)

        v = relkin.get_speed(E, self.particle.m)       # m/s
        v_par = v*pitch
        v_perp = v*np.sqrt(1 - pitch**2)

        return v_par, v_perp

    def sample_energy_pitch(self, n, index=0):
        """Sample `n` energy and pitch values from the distribution, to be overloaded by the sub-classes."""
        n = int(n)
        return np.zeros(n), np.zeros(n)


class SpeedDistribution(VparVperpDistribution):
    """A velocity distribution characterized mainly by its distribution in speed.

    The speeds are determined from an arbitrary speed distribution and the 
    direction of the corresponding velocity vectors are taken to be uniformly 
    distributed in a given pitch range, where 

        pitch = v_parallel/v

    and v_parallel is the component of the velocity relative to a given 
    reference direction.

    Typically, a user would subclass this class and override the `sample_speed` method.

    Attributes
    ----------
    pitch_range : array-like with 2 elements
        Pitch values of the particles are taken to be uniformly distributed 
        between the two values in this sequence.

    For the rest of the attributes see docstring of the parent class(es)."""

    def __init__(self, particle, density=None, v_collective=None, pitch_range=[-1,1], ref_dir=[0,1,0]):

        super().__init__(particle, density=density, v_collective=v_collective, ref_dir=ref_dir)
        self.pitch_range = pitch_range

    def sample_vpar_vperp(self, n, index=0):
        v = self.sample_speed(n, index=index)
        pitch = self.sample_pitch(n)

        v_par = v*pitch
        v_perp = v*np.sqrt(1 - pitch**2)

        return v_par, v_perp

    def sample_speed(self, n, index=0):
        """Sample `n` speed values from the distribution, to be overloaded by the sub-classes."""
        n = int(n)
        return np.zeros(n)

    def sample_pitch(self, n):
        """Sample `n` pitch values uniformly distributed in the range given by self.pitch_range."""
        return sampler.sample_uniform(self.pitch_range, n)


class EnergyDistribution(EnergyPitchDistribution):
    """A velocity distribution characterized mainly by its distribution in energy.

    The speeds are determined from an arbitrary energy distribution and the 
    direction of the corresponding velocity vectors are taken to be uniformly 
    distributed in a given pitch range, where 

        pitch = v_parallel/v

    and v_parallel is the component of the velocity relative to a given 
    reference direction.

    Typically, a user would subclass this class and override the `sample_energy` method.

    Attributes
    ----------
    pitch_range : array-like with 2 elements
        Pitch values of the particles are taken to be uniformly distributed 
        between the two values in this sequence.

    For the rest of the attributes see docstring of the parent class(es)."""


    def __init__(self, particle, density=None, v_collective=None, pitch_range=[-1,1], ref_dir=[0,1,0]):

        super().__init__(particle, density=density, v_collective=v_collective, ref_dir=ref_dir)
        self.pitch_range = pitch_range

    def sample_energy_pitch(self, n, index=0):
        E = self.sample_energy(n, index=index)
        p = self.sample_pitch(n)

        return E, p

    def sample_energy(self, n, index=0):
        """Sample `n` energy values from the distribution, to be overloaded by the sub-classes."""
        n = int(n)
        return np.zeros(n)

    def sample_pitch(self, n):
        """Sample `n` pitch values uniformly distributed in the range given by self.pitch_range."""
        return sampler.sample_uniform(self.pitch_range, n)
        

class MaxwellianDistribution(EnergyDistribution):
    """A class representing a Maxwellian velocity distribution.

    Attributes
    ----------    
    T : float or array of shape (N,)
        Temperature (in keV).

    For the rest of the attributes see docstring of the parent class(es)."""

    
    def __init__(self, T, particle, density=None, v_collective=None, pitch_range=[-1,1], ref_dir=[0,1,0]):

        super().__init__(particle, density=density, v_collective=v_collective, 
                         pitch_range=pitch_range, ref_dir=ref_dir)
        
        self.T = np.atleast_1d(T)
        self.n_spatial = len(self.T)

    def sample_energy(self, n, index=0):
        """Sample energies (in keV) from the Maxwellian distribution."""
        
        # The (kinetic) energy is distributed as a chi2 variable with 3 degrees of freedom
        E = np.random.chisquare(3, size=n) * 0.5 * self.T[index]
        
        return E
        

class MonoEnergeticDistribution(EnergyDistribution):
    """A class representing a mono-energetic velocity distribution.

    Attributes
    ----------
    E0 : float or array of length N
        Kinetic energy of the particles (in keV)

    For the rest of the attributes see docstring of the parent class(es)."""

    def __init__(self, E0, particle, density=None, pitch_range=[-1,1], ref_dir=[0,1,0]):
        
        super().__init__(particle, density=density, v_collective=None, 
                         pitch_range=pitch_range, ref_dir=ref_dir)
        
        self.E0 = np.atleast_1d(E0)

    def sample_energy(self, n, index=0):
        """Sample energies (in keV) from the mono-energetic distribution."""
        return sampler.sample_mono(self.E0[index], n)


class TabulatedEnergyDistribution(EnergyDistribution):
    """A class representing a velocity distribution where the energy distribution
    is given by tabulated values of f vs E.

    Attributes
    ----------
    E_axis : array of shape (N,)
        The energy values (keV) of the tabulated distribution.
    
    dist : array of shape (N,)
        The distribution value at the tabulated energies. The absolute normalization 
        of the distribution is not important, but the values f(E)*dE should be proportional
        to the number of particles in the energy interval dE.

    For the rest of the attributes see docstring of the parent class(es)."""

    def __init__(self, E_axis, dist, particle, density=None, pitch_range=[-1,1], ref_dir=[0,1,0]):
        
        super().__init__(particle, density=density, v_collective=None, 
                         pitch_range=pitch_range, ref_dir=ref_dir)

        self.E_axis = np.array(E_axis)
        self._set_1d_dist(dist)
    
    def sample_energy(self, n, index=0):
        """Sample energies (in keV) from tabulated energy distribution."""
        
        n = int(n)
        E = sampler.sample_tab(self.dist[index], self.E_axis, n_samples=n)

        return E


class TabulatedEnergyPitchDistribution(EnergyPitchDistribution):
    """A class representing a velocity distribution where the energy-pitch distribution
    is given by tabulated values of f(E,pitch).

    Attributes
    ----------
    E_axis : array of shape (NE,)
        The energy values (keV) of the tabulated distribution.

    pitch_axis : array of shape (NP,)
        The pitch values of the tabulated distribution

    dist : array of shape (NE,NP)
        The distribution value at the tabulated (energy,pitch)-values. The absolute normalization 
        of the distribution is not important, but the values f(E,p)*dE*dp should be proportional
        to the number of particles in the phase space volume dE*dp.

    dE : array of shape (NE,)
        Width of the energy bins (keV). If not given, the width will be inferred from `E_axis`
        when sampling the distribution.

    d_pitch : array of shape (NP,)
        Width of the pitch bins. If not given, the width will be inferred from `E_axis`
        when sampling the distribution.

    For the rest of the attributes see docstring of the parent class(es)."""

    def __init__(self, E_axis, pitch_axis, dist, particle, density=None, dE=None, d_pitch=None, ref_dir=[0,1,0]):
        """Create tabulated energy-pitch distributions from given data.

        Parameters
        ----------
        E_axis : array-like of length NE
            The energy values (keV) of the tabulated distribution.

        pitch_axis : array-like of length NP
            The pitch values of the tabulated distribution

        dist : array-like of shape (NE,NP)
            The distribution value at the tabulated (energy,pitch)-values. The absolute normalization 
            of the distribution is not important, but the values f(E,p)*dE*dp should be proportional
            to the number of particles in the phase space volume dE*dp.

        dE : array of shape (NE,)
            Width of the energy bins (keV). If not given, the width will be inferred from `E_axis`
            when sampling the distribution.

        d_pitch : array of shape (NP,)
            Width of the pitch bins. If not given, the width will be inferred from `pitch_axis`
            when sampling the distribution."""

        super().__init__(particle, density=density, v_collective=None, ref_dir=ref_dir)
        
        self.E_axis = np.array(E_axis)
        self.pitch_axis = np.array(pitch_axis)
        self._set_2d_dist(dist)

        self.dE = dE
        self.d_pitch = d_pitch

    def sample_energy_pitch(self, n, index=0):
        """Sample energies (keV) and pitch values from the tabluated distribution."""
        
        sample = sampler.sample_tab(self.dist[index], self.E_axis, self.pitch_axis, 
                                    dx=[self.dE, self.d_pitch], n_samples=n)

        E = sample[0]
        pitch = sample[1]

        return E, pitch


class TabulatedVparVperpDistribution(VparVperpDistribution):
    """Velocity distribution represented by tabulated values of f(v_par, v_perp).

    Attributes
    ----------
    v_par_axis : array of shape (Nvpa,)
        The v_par axis of the tabulated distribution (m/s).

    v_perp_axis : array of shape (Nvpe,)
        The v_perp values of the tabulated distribution (m/s).

    dist : array of shape (Nvpa,Nvpe) or (N,Nvpa,Nvpe)
        The distribution value at the tabulated (v_par,v_perp)-values. The absolute normalization 
        of the distribution is not important, but the values f(v_par,v_perp)*dv_par*dv_perp should be proportional
        to the number of particles in the phase space volume dv_par*dv_perp.

    dv_par : array of shape (Nvpa,)
        Width of the v_par bins (m/s).

    dv_perp : array of shape (Nvpe,)
        Width of the v_perp bins (m/s).

    For the rest of the attributes see docstring of the parent class(es)."""

    def __init__(self, v_par_axis, v_perp_axis, dist, particle, density=None, dv_par=None, dv_perp=None, ref_dir=[0,1,0]):

        super().__init__(particle, density=density, v_collective=None, ref_dir=ref_dir)
        
        self.v_par_axis = np.array(v_par_axis)
        self.v_perp_axis = np.array(v_perp_axis)
        self._set_2d_dist(dist)

        self.dv_par = dv_par
        self.dv_perp = dv_perp

    def sample_vpar_vperp(self, n, index=0):
        """Sample parallel and perpendicular speeds (m/s) from the tabluated distribution."""
        
        sample = sampler.sample_tab(self.dist[index], self.v_par_axis, self.v_perp_axis, 
                                    dx=[self.dv_par, self.dv_perp], n_samples=n)

        v_par = sample[0]
        v_perp = sample[1]
        
        return v_par, v_perp


class TabulatedSpeedDistribution(SpeedDistribution):
    """A class representing a velocity distribution where the speed distribution
    is given by tabulated values of f vs v.

    Attributes
    ----------
    v_axis : array of shape (N,)
        The speed values (m/s) of the tabulated distribution.
    
    dist : array of shape (N,)
        The distribution value at the tabulated speeds. The absolute normalization 
        of the distribution is not important, but the values f(v)*dv should be proportional
        to the number of particles in the energy interval dE.

    For the rest of the attributes see docstring of the parent class(es)."""

    def __init__(self, v_axis, dist, particle, density=None, pitch_range=[-1,1], ref_dir=[0,1,0]):
        
        super().__init__(particle, density=density, v_collective=None, 
                         pitch_range=pitch_range, ref_dir=ref_dir)

        self.v_axis = np.array(v_axis)
        self._set_1d_dist(dist)
    
    def sample_speed(self, n, index=0):
        """Sample speeds (in m/s) from tabulated speed distribution."""
        
        n = int(n)
        v = sampler.sample_tab(self.dist[index], self.v_axis, n_samples=n)

        return v



"""Collection of 3-vector operations that are required by the `dress` package.

A 3-vector is represented by an array with shape (3,1). Similarly, N vectors
can be represented by an array of shape (3,N)."""

import numpy as np


def make_vector(*vecs):
    """Convert given arrays (or array-likes) to dress-vectors, if possible."""
    
    new_vecs = []

    for vec in vecs:
        v = np.atleast_2d(vec)
        
        # If input represented one single vector the array needs to be 
        # transposed in order to have size (3,1)
        if v.shape[0] == 1:
            v = v.T

        new_vecs.append(v)

    # Check that everything worked
    check_vector(*new_vecs)

    if len(new_vecs) == 1:
        return new_vecs[0]
    else:
        return new_vecs


def check_vector(*vecs):
    """Check that the given arrays represent valid 3-vectors."""

    for vec in vecs:
        if not isinstance(vec, np.ndarray):
            raise TypeError('Vector must be a numpy array with shape (3,N)')

        if not (vec.ndim == 2 and vec.shape[0] == 3):
            msg = f'\n{vec}\n does not represent valid array of three-vectors (shape must be (3,N))'
            raise ValueError(msg)


def dot(v1, v2):
    """Dot product between two vectors.

    Parameters
    ----------
    v1, v2 : array of shape (3,N) or (3,1)
        The vectors to multiply.

    Returns
    -------
    prod : array of shape (N,)
        The dot product of the vectors."""
    
    check_vector(v1, v2)
    prod = np.sum(v1*v2, axis=0)
    
    return prod


def cross(v1, v2):
    """Cross product between two vectors.

    Parameters
    ----------
    v1, v2 : array of shape (3,N)
        The vectors to multiply.

    Returns
    -------
    prod : array of shape (N,)
        The cross product of the vectors."""

    check_vector(v1, v2)
    prod = np.cross(v1, v2, axis=0)
    
    return prod
    
    
def normalize(v, length=1):
    """Normalize vectors to given length.

    Parameters
    ----------
    v : array of shape (3,N)
        The vectors to be normalized.
    
    length : scalar or array of shape (N,)
        The new lengths of the vectors

    Returns
    -------
    u : array of shape (3,N)
        The renormalized vectors."""
    
    check_vector(v)
    
    N = v.shape[1]
    length = np.atleast_1d(length)
    if length.shape != (1,) and length.shape != (N,):
        raise ValueError(f'`length` must have either 1 element or {N} element')

    old_length = np.sqrt(dot(v,v))
    u = v * length / old_length

    return u


def repeat(v, n):
    """Repeat the vector `v` N times.

    Parameters
    ----------
    v : array-like of length 3
        The vector to be repeated.
    
    n : int
        Number of times to repeat.

    Return
    ------
    vn : array of shape (3,n)
        The vector `v` repeated n times."""
    
    v = make_vector(v)
    
    if v.shape != (3,1):
        raise ValueError('`v` cannot be interpreted as a (single) vector')

    vn = np.repeat(v, n, axis=1)

    return vn



"""Module for relativistic kinematics calculations."""

import numpy as np
import scipy.constants as const


def get_speed(E, m):
    """
    Calculate the speed (in m/s) of a particle with kinetic energy E (keV)
    and mass 'm' (keV/c**2)."""

    E = E + m                              # total energy (keV)
    v = np.sqrt(1 - (m/E)**2)*const.c      # speed (m/s)

    return v


def get_energy(v, m):
    """
    Calculate the kinetic energy 'E' (in keV) of a particle with velocity 'v' (m/s)
    and mass 'm' (keV/c**2).

    'v' can be a length-N array with the values of the speed, or a (3,N) array
    with the velocity components.

    Relativistic kinematics is used."""

    v = np.atleast_1d(v)

    c = const.c         # speed of light (m/s)

    v = np.linalg.norm(v, ord=2, axis=0)
    gamma = 1/np.sqrt(1-(v/c)**2)

    E = (gamma-1)*m     # keV

    return E


def four_vector(E, p):
    E = np.atleast_1d(E)

    if p.ndim == 1:
        p = p.reshape(3,1)

    return np.vstack((E, p))


def get_four_momentum(vel, m):
    """
    Calculate the four-momentum (keV/c) for a particle with velocity 'vel' (m/s)
    and mass 'm' (keV/c**2).

    'vel' can also be a (3,N) array with velocities and 'm' can be a
    length-N array with masses.
    """

    vel = vel / const.c

    v2 = vel[0]**2 + vel[1]**2 + vel[2]**2

    gamma = 1/np.sqrt(1 - v2)

    E = gamma*m                           # total energy (keV)
    p = gamma*m*vel                       # three-momentum (keV/c)

    P = four_vector(E, p)

    return P


def mult_four_vectors(P1, P2):
    """
    Multiply two 4-vectors.
    Vectorized input works if 'P1' and/or 'P2' are (4,N) arrays.
    """

    prod = P1[0]*P2[0] - np.sum(P1[1:]*P2[1:], axis=0)

    return prod


def boost(P, beta):
    """
    Lorentz transformation of four-momentum P.

    All vector input should be 2D, i.e. shape (4,N) for four-vectors
    and (3,N) for three-vectors, where N is either equal to 1 or to
    the number of particles.
    """

    # Evaluate three-momentum parallel and perpendicular to beta
    E = P[0]
    p3 = P[1:]

    beta_mag = np.linalg.norm(beta, axis=0)
    i_nonzero = beta_mag != 0

    beta_dir = np.zeros_like(beta)
    beta_dir[:,i_nonzero] = beta[:,i_nonzero] / beta_mag[i_nonzero]  # avoid division by zero

    p3_par_mag = np.sum(p3*beta_dir, axis=0)
    p3_par = p3_par_mag * beta_dir
    p3_perp = p3 - p3_par

    # Gamma factor
    gamma = 1 / np.sqrt(1-beta_mag**2)

    # Lorentz boost
    E_prime = gamma*E - gamma*beta_mag*p3_par_mag
    p3_par_mag_prime = -gamma*beta_mag*E + gamma*p3_par_mag     # Only the parallel component of p3 is affected

    p3_prime = p3_par_mag_prime * beta_dir + p3_perp

    # Final result
    P_prime = four_vector(E_prime, p3_prime)

    return P_prime



import os

import numpy as np

from dress.config import cross_section_dir
from dress.reactions.particle import Particle
from dress.reactions.reaction import Reaction


class TT2NHe4Reaction(Reaction):
    """Class representing the t + t -> n + n + he4 fusion reaction."""

    __slots__ = ('tab_cross_section')

    def __init__(self):

        super().__init__('t', 't', 'n', 'n', 'he4')

        # Load tabulated cross section (from ENDF)
        cx_path = os.path.join(cross_section_dir,'T(t,2n)He4-cross-section.txt')
        tab_cross_section = np.loadtxt(cx_path, usecols=[0,1], comments='#')
        
        tab_cross_section[:,0] /= 2000.0   # convert from lab to CM, and from eV to keV
        tab_cross_section[:,1] *= 1e-28    # convert from barn to m**2

        self.tab_cross_section = tab_cross_section
        
    
    def calc_sigma_tot(self, E):
        """Evaluate total cross section.

        Arguments
        ----------
        E : array-like
            Reactant energies in the center of mass frame (keV).

        Returns
        -------
        array
            The cross section (in m**2) for each energy value.
        """
        
        sigma =  np.interp(E, self.tab_cross_section[:,0], self.tab_cross_section[:,1],
                           left=0.0, right=0.0)

        return sigma
        

    def calc_sigma_diff(self, E, costheta):
        """Evaluate the angular differential cross section.

        Arguments
        ---------
        E : array-like
            Reactant energies in the center of mass frame (keV).
        costheta : array-like
            Cosine of the emission angle of the neutron 
            (with respect to the reactant relative velocity)

        Returns
        -------
        array
            The cross section (in m**2 sr**-1) for each energy value.
        """

        E = np.atleast_1d(E)
        costheta = np.atleast_1d(costheta)

        # Assume isotropic cross section
        sigma = self.calc_sigma_tot(E)/(4*np.pi)    # m**2/sr

        return sigma



"""Collection of useful functions for working with particles
moving around in a tokamak."""

import numpy as np
import scipy.constants as const

from dress.relkin import get_speed, get_energy


def get_vpar_vperp(E, p, m):
    """
    Calculate velocity components (in m/s) parallel and perpendicular
    to the plasma magnetic field, for a particle of mass 'm' (keV/c**2)
    with given values of energy 'E' (keV) and pitch p = v_par/v.
    """

    v = get_speed(E, m)

    v_par  = v*p
    v_perp = v*np.sqrt(1-p**2)

    return v_par, v_perp


def get_EP(v_par, v_perp, m):
    """
    Calculate energy E (keV) and pitch p = v_par/v
    for given values of the velocity components (in m/s)
    parallel and perpendicular to the plasma magnetic field,
    for a particle with mass 'm' (keV/c**2).
    """

    v = np.sqrt(v_par**2 + v_perp**2)

    E = get_energy(v, m)
    p = v_par/v

    return E, p


def get_basis_vectors(B):
    """
    Construct a set of orthogonal basis vectors that aligns with
    a given magnetic field vector 'B' (or a (3,N) array of N vectors).

    The components of the basis vectors are evaluated in cylindrical coordinates.

    (The unit of 'B' is not important, only its direction is used by the function.)
    """

    if np.ndim(B) == 1:
        B = B.reshape(3,1)    # reshape needed for broadcasting to work

    # Normalize the B-field vector
    B_norm = np.linalg.norm(B, ord=2, axis=0)
    b = B/B_norm

    # Radial basis vector in cylindrical coordinates
    R_hat   = np.array([1,0,0]).reshape(3,1)    # reshape needed for broadcasting to work

    # Construct basis vectors that align with B
    # (only works if the phi component of B is non-zero)
    e1 = R_hat - b[0]*b
    e1 = e1 / np.linalg.norm(e1, ord=2, axis=0)
    e2 = b
    e3 = np.cross(e1,e2, axis=0)

    return e1, e2, e3


def sample_gyro_angle(N):
    """Sample 'N' random gyro angles in the interval [0,2*pi)."""

    return 2*np.pi*np.random.rand(N)


def get_velocity(v_par, v_perp, e1, e2, e3, theta_g):
    """
    Calculate the velocity components (m/s) for given values of
    the speeds parallel and perpendicular to the plasma magnetic field.

    'e1', 'e2' and 'e3' are orthogonal basis vectors, such that 'e2' is
    parallel to the magnetic field vector.

    'theta_g' is the gyro radius.
    """

    v_par = v_par * e2
    v_perp = v_perp*(-np.sin(theta_g)*e1 + np.cos(theta_g)*e3)

    return v_par + v_perp


def get_Larmor_radius(v_perp, B, Z, m):
    """
    Calculate the Larmor radius for a particle with speed v_perp (m/s)
    relative to the plasma magnetic field 'B' (T).

    'v_perp' and 'B' can be length-N arrays with the relevant magnitudes.
    Alternatively, B could also be a (3,N) array with vector components.

    'Z' and 'm' are the atomic number and mass (keV/c**2) of the particle(s).
    """

    B = np.atleast_2d(B)

    q = Z*const.e

    B = np.linalg.norm(B, ord=2, axis=0)

    m = m * 1000 * const.e / const.c**2     # keV/c**2 -> kg
    r_L = m*v_perp / (q*B)

    return abs(r_L)


def add_gyration(E, p, m, B, theta_g=None, flr=False, Z=None, Rg=None, Zg=None):
    """
    Calculate the cylindrical velocity vector of a particle in a tokamak.
    'theta_g' is the gyro angle; if 'None' it is sampled randomly from [0,2*pi).

    'E' is the particle kinetic energy (keV). 'p' = v_par/v is the pitch.
    'm' is the mass (keV/c**2).
    'B' is the magnetic field vector at the gyrocenter of the particle (T).

    If 'flr' = True, the function also returns the position of the particle
    (cylindrical coordinates). In this case, the gyrocenter position ('Rg','Zg')
    needs to be provided (in meters), as well as the atomic mass number 'Z'.

    Vectorized input works if 'B' is given as a (3,N) array and all other
    input are length-N arrays (or scalars).
    """

    if np.ndim(B) == 1:
        B = B.reshape(3,1)    # reshape needed for broadcasting to work

    E = np.atleast_1d(E)

    # Local coordinate system that aligns with B
    e1, e2, e3 = get_basis_vectors(B)

    # Compute the velocity vector in the local coordinate system.
    v_par, v_perp = get_vpar_vperp(E, p, m)    # speeds

    if theta_g == None:
        theta_g = sample_gyro_angle(len(E))

    v = get_velocity(v_par, v_perp, e1, e2, e3, theta_g)    # velocity vectors

    # Compute particle position, if required
    if flr:
        # Larmor radius
        r_L = get_Larmor_radius(v_perp, B, Z, m)

        # The gyro-center is at (Rg,phi_g,Zg). Actual position should be
        # displaced by one Larmor radius in the (e1,e3) plane.
        R_hat = np.array([1,0,0]).reshape(3,1)    # reshape needed for broadcasting to work
        Z_hat = np.array([0,0,1]).reshape(3,1)

        x = Rg*R_hat + Zg*Z_hat + r_L*(np.cos(theta_g)*e1 + np.sin(theta_g)*e3)

        return v, x

    else:
        return v



"""Useful tools for performing reaction product spectrum calculations 
for tokamak plasmas."""

import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import griddata, interp1d

import dress.utils


class FluxSurfaceMap:
    """Mapping between the spatial coordinates (R,Z) and the 
    flux surface label rho."""

    def __init__(self, R, Z, rho, rho_lim=1.0, rho_outside=2.0):
        """Construct a flux surface map rho = f(R,Z) from given tabulated data.

        Parameters
        ----------
        R : array-like of shape (NP,)
            R coordinates of the tabulated flux surfaces.

        Z : array-like of shape (NP,)
            Z coordinates of the tabulated flux surfaces.
        
        rho : array-like of shape (NP,)
            Flux coordinate value at each of the given (R,Z) points.

        rho_lim : float
            Flux surface label of the last closed flux surface.
            Default is 1.0.

        rho_outside : float
            Fill value to use outside the last closed flux surface.
            Default is 2.0."""

        R = np.atleast_1d(R)
        Z = np.atleast_1d(Z)
        rho = np.atleast_1d(rho)

        if not (1 == R.ndim == Z.ndim == rho.ndim):
            raise ValueError('Input data must be 1D arrays of tabulated flux values')

        if not (len(R) == len(Z) == len(rho)):
            raise ValueError('Input data must have the same length')

        self.Rtab = R
        self.Ztab = Z
        self.rho_tab = rho

        self.rho_lim = rho_lim
        self.rho_outside = rho_outside


    def get_rho(self, R, Z):
        """Evaluate rho for given (R,Z) values.

        Parameters
        ----------
        R, Z : 1D array-likes of same length
            The (R,Z) values where rho is to be evaluated.

        Returns
        -------
        rho : 1D array
            The flux surface label at the requested points."""

        rho = griddata((self.Rtab, self.Ztab), self.rho_tab, (R, Z), method='linear', 
                       fill_value=self.rho_outside)

        return rho


    def get_closest_gridpoint(self, R, Z):
        """Evaluate closest grid point of the given (R,Z) position.

        Parameters
        ----------
        R, Z : 1D array-likes of same length
            The (R,Z) values for which to find the closest grid points.

        Returns
        -------
        grid_index : array of int's
            Indices of the closest grid points."""

        index_values = np.arange(len(self.rho_tab))
        rho_index = griddata((self.Rtab, self.Ztab), index_values, (R, Z), method='nearest')

        return rho_index.astype('int')

        
    def calc_dvol(self, rho_axis):
        """Get volume enclosed between each flux surface pairs in `rho_axis`."""

        # Generate rectangular grid
        R = np.linspace(self.Rtab.min(), self.Rtab.max(), 250)
        Z = np.linspace(self.Ztab.min(), self.Ztab.max(), 300)
        dR = R[1] - R[0]
        dZ = Z[1] - Z[0]
        R, Z = np.meshgrid(R, Z, indexing='ij')

        R = R.flatten()
        Z = Z.flatten()
        dV = 2*np.pi*R*dR*dZ
        rho_grid = self.get_rho(R, Z)

        # Sum volume of grid points inside each flux surface
        dvol = np.zeros_like(rho_axis)
        vol_tot = 0.0

        for i in range(len(rho_axis)):
            enclosed = rho_grid <= rho_axis[i]
            dvol[i] = np.sum(dV[enclosed]) - vol_tot
            vol_tot += dvol[i]

        return dvol


class FluxSurfaceQuantity:
    """A class for representing flux surface quantities in tokamaks."""

    def __init__(self, rho, val, flux_surface_map):
        """Construct a flux surface quantity from tabulated values.

        Parameters
        ----------
        rho, val : 1D array-likes
            The flux surface values and corresponding values of the tabulated profile.

        flux_surface_map : FluxSurfaceMap
            Mapping between (R,Z) and rho."""

        
        rho = np.atleast_1d(rho)
        val = np.atleast_1d(val)

        if not (1 == rho.ndim == val.ndim):
            raise ValueError('Input data must be 1D arrays of tabulated values')

        if not (len(rho) == len(val)):
            raise ValueError('Input data must have the same length')

        if not isinstance(flux_surface_map, FluxSurfaceMap):
            raise ValueError(f'`flux_surface_map` must be an instance of {FluxSurfaceMap}')

        self.flux_surface_map = flux_surface_map
        self.rho = rho
        self.val = val

        self.eval_from_rho = interp1d(rho, val, kind='linear', fill_value=(val[0],0), bounds_error=False)
        

    def eval_from_RZ(self, R, Z):
        
        rho = self.flux_surface_map.get_rho(R, Z)
        val = self.eval_from_rho(rho)

        return val


def get_toroidal_velocity(angular_freq, R, Z, toroidal_unit_vector=[0,1,0]):
    """Evaluate toroidal velocity at the given (R,Z) points.

    Parameters
    ----------
    angular_freq : instance of dress.tokamak.utils.FluxSurfaceQuantity
        The angular frequency profile (radians/sec).

    R, Z : array-likes of shape (NP,)
        The positions where the toroidal rotation is to be evaluated.

    toroidal_unit_vector : array-like of shape (3,) of (NP,3)
        Components of the toroidal unit vector in the reference frame of interest.
        If 1D array, it is assumed to be the same for all the requested points.

    Returns
    -------
    v_tor : array of shape (NP,3)
        Toroidal velocities at each of the requested points."""

    R, Z = np.atleast_1d(R, Z)
    angf = angular_freq.eval_from_RZ(R, Z)
    
    toroidal_speed = angf*R     # m/s
    
    toroidal_unit_vector = np.atleast_2d(toroidal_unit_vector)
    NP = len(toroidal_speed)

    if toroidal_unit_vector.shape == (1,3):
        toroidal_unit_vector = np.repeat(toroidal_unit_vector, NP, axis=0)

    v_tor = toroidal_speed[:,None] * toroidal_unit_vector

    return v_tor



"""Particle masses used by the 'reactions' module."""

from scipy.constants import physical_constants as const


# Atomic mass unit
u_keV = const['atomic mass constant energy equivalent in MeV'][0] * 1000.0      # keV
u_kg  = const['atomic mass constant'][0]   # kg

# Electron mass
me = const['electron mass in u'][0]

# Neutron mass
mn = const['neutron mass in u'][0]

# Proton mass
mp = const['proton mass in u'][0]

# Deuteron mass
md = const['deuteron mass in u'][0]

# Triton mass
mt = const['triton mass in u'][0]

# 3-He mass
# (from https://www.nndc.bnl.gov/nudat3/. Atomic binding energy neglected)
m3He = (3*u_keV + 14931.21888) / u_keV
m3He = m3He - 2*me        # u

# 4-He mass
m4He = const['alpha particle mass in u'][0]

# 6-Li mass
# (from https://www.nndc.bnl.gov/nudat3/. Atomic binding energy neglected)
m6Li = (6*u_keV + 14086.8804) / u_keV
m6Li = m6Li - 3*me        # u

# 8-Be mass
# (from https://www.nndc.bnl.gov/nudat3/. Atomic binding energy neglected)
m8Be = (8*u_keV + 4941.67) / u_keV
m8Be = m8Be - 4*me        # u

# 9-Be mass
# (from https://www.nndc.bnl.gov/nudat3/. Atomic binding energy neglected)
m9Be = (9*u_keV + 11348.45) / u_keV
m9Be = m9Be - 4*me        # u

# 9-B mass
# (https://www.nndc.bnl.gov/nudat3/. Atomic binding energy neglected)
m9B = (9*u_keV + 12416.5) / u_keV
m9B = m9B - 5*me          # u

# 10-B mass
# (https://www.nndc.bnl.gov/nudat3/. Atomic binding energy neglected)
m10B = (10*u_keV + 12050.611) / u_keV
m10B = m10B - 5*me          # u

# 12-C mass
# (from https://www.nndc.bnl.gov/nudat3/. Atomic binding energy neglected)
m12C = 12*u_keV / u_keV
m12C = m12C - 6*me        # u




import numpy as np

from dress.reactions import masses


class Particle:
    """A class for holding particle info."""

    __slots__ = ('name', 'long_name', 'u', 'excitation_energy', 'm')

    def __init__(self, name):

        name = name.lower()

        self.excitation_energy = 0.0

        if name == 'n':
            self.name = 'n'
            self.long_name = 'neutron'
            self.u = masses.mn

        elif name in ['p', 'h']:
            self.name = 'p'
            self.long_name = 'proton'
            self.u = masses.mp

        elif name == 'd':
            self.name = 'd'
            self.long_name = 'deuteron'
            self.u = masses.md

        elif name == 't':
            self.name = 't'
            self.long_name = 'triton'
            self.u = masses.mt

        elif name in ['3he', 'he3']:
            self.name = 'he3'
            self.long_name = 'helium-3'
            self.u = masses.m3He

        elif name in ['4he', 'he4', 'alpha']:
            self.name = 'he4'
            self.long_name = 'helium-4'
            self.u = masses.m4He

        elif name in ['6li', 'li6']:
            self.name = 'li6'
            self.long_name = 'lithium-6'
            self.u = masses.m6Li

        elif name in ['8be', 'be8']:
            self.name = 'be8'
            self.long_name = 'beryllium-8'
            self.u = masses.m8Be

        elif name in ['9be', 'be9']:
            self.name = 'be9'
            self.long_name = 'beryllium-9'
            self.u = masses.m9Be

        elif name in ['9b', 'b9']:
            self.name = 'b9'
            self.long_name = 'boron-9'
            self.u = masses.m9B

        elif name in ['10b', 'b10']:
            self.name = 'b10'
            self.long_name = 'boron-10'
            self.u = masses.m10B

        elif name in ['12c', 'c12']:
            self.name = 'c12'
            self.long_name = 'carbon-12'
            self.u = masses.m12C

        elif name in ['12c(e1)', 'c12(e1)']:
            self.name = 'c12(e1)'
            self.long_name = 'carbon-12 (1st excited state)'
            self.u = masses.m12C
            
            # 1st excited state according to nndc.bnl.gov
            self.excitation_energy = 4439.8    # keV
            
        else:
            raise ValueError(f'Invalid particle name: {name}')

        self.m = self.u * masses.u_keV + self.excitation_energy


    def __repr__(self):
        return 'Particle: {}'.format(self.long_name)

        
    def __eq__(self, other):
        
        if isinstance(other, Particle):
            if self.long_name == other.long_name:
                return True
            else:
                return False

        else:
            return False



"""Classes for calculating neutron spectra using the
'sampler' and 'relscatt' modules.
"""

import numpy as np
from scipy.interpolate import interp1d, RectBivariateSpline

from dress import relkin
from dress import relscatt
from dress import sampler
from dress import vec_ops as vec


class Reactant:
    """Class for holding reactant info and samples."""
    
    def __init__(self, particle, n_samples=1e6):
        
        self.particle = particle

        # Initialize without MC sample
        self.P = None
        self.n_samples = int(n_samples)

    def __repr__(self):
        return 'Reactant: {}'.format(self.particle.long_name)

    @property
    def m(self):
        return self.particle.m

    # Handling the sample vectors
    @property
    def v(self):
        return self._v

    @v.setter
    def v(self, v_):
        if v_.shape != (3,self.n_samples):
            raise ValueError('Wrong shape of v (wrong number of samples?)')
        else:
            self._v = v_
            self.P = relkin.get_four_momentum(v_, self.m)
        

# Calculation of spectrum from sampled four-momenta.
# -----------------------------------------------------

class SpectrumCalculator:
    """ 
    Calculate product spectrum from reactions between particles with
    four-momenta 'Pa' and 'Pb'.
    """

    def __init__(self, reaction, n_samples=1e6, ref_dir=None):

        self._n_samples = int(n_samples)
        self.reaction = reaction
        self.weights = None
        self.ref_dir = ref_dir

        # 4*pi emission by default
        self.u = None

    def __repr__(self):
        return f'Spectrum calculator for the reaction {self.reaction.formula}'

    @property
    def reaction(self):
        return self._reaction

    @reaction.setter
    def reaction(self, r):
        
        self._reaction = r

        self.reactant_a = r.a
        self.reactant_b = r.b
        self.product_1 = r.p1
        self.product_2 = r.p2
        self.product_3 = r.p3

    @property
    def reactant_a(self):
        return self._reactant_a

    @reactant_a.setter
    def reactant_a(self, particle):
        self._reactant_a = Reactant(particle, n_samples=self.n_samples)

    @property
    def reactant_b(self):
        return self._reactant_b

    @reactant_b.setter
    def reactant_b(self, particle):
        self._reactant_b = Reactant(particle, n_samples=self.n_samples)

    @property
    def ma(self):
        return self.reactant_a.m

    @property
    def mb(self):
        return self.reactant_b.m

    @property
    def m1(self):
        return self.product_1.m

    @property
    def m2(self):
        return self.product_2.m

    @property
    def m3(self):
        if self.product_3 is None:
            return None
        else:
            return self.product_3.m
    
    @property
    def n_samples(self):
        return self._n_samples

    @n_samples.setter
    def n_samples(self, n):
        n = int(n)
        self._n_samples = n
        self.reactant_a.n_samples = n
        self.reactant_b.n_samples = n
        
    @property
    def weights(self):
        if self._weights is None:
            return np.ones(self.n_samples)/self.n_samples
        else:
            return self._weights

    @weights.setter
    def weights(self, w):
        self._weights = w

    @property
    def ref_dir(self):
        return self._ref_dir

    @ref_dir.setter
    def ref_dir(self, u):
        
        if u is None:
            self._ref_dir = None
        else:
            u = vec.make_vector(u)
            u = vec.normalize(u)
            self._ref_dir = u
    

    def __call__(self, bins=None, bin_width=25.0, normalize=False, which_product=1):
        """
        Compute reactant spectrum. The units of the spectrum depends on the 
        units of the 'weights' attribute. The 'normalize' keyword
        can be used to normalize the spectrum to unit sum.
    
        If 'bins' is given it is passed as a keyword argument to np.histogram. 
        If bins=None (default), the bins will be constructed so as to cover the full 
        range of calculated product energies.
        
        The bin width of the spectrum histogram can be specified with 'bin_width' (keV). 
        (only effective if bins=None).
        
        Additionally, both the 'bins' and the 'bin_width' attributes can be length-2 lists,
        in which case the spectrum is resolved both in energy and cos(theta), where
        theta is the angle between the B-field and the emission direction.
        """        
        
        # Calculate relevant kinematic data of the reactants
        Pa = self.reactant_a.P
        Pb = self.reactant_b.P

        reactant_data = relscatt.ReactantData(Pa, Pb, self.ma, self.mb)

        # Emission directions
        if self.u is None:
            u = sampler.sample_sphere(self.n_samples)   # random emission directions
        else:
            u = np.array(self.u)
            if u.ndim == 1:
                u = np.array(u).reshape(3,1)   # same emission direction for all particles            

        # Setup for computing the spectrum of the requested product
        m, mr, mr2 = self._get_product_masses(which_product)

        # Compute product four-momenta
        if self.product_3 is not None:             
            P = relscatt.three_body_event(reactant_data, m, mr, mr2, u)
        else:
            P = relscatt.two_body_event(reactant_data, m, mr, u)

        # Compute reactivity
        sigmav = relscatt.get_reactivity(reactant_data, P, self.reaction)
    
        # Make attributes with reactivity for possible later use
        self.sigmav = sigmav

        # Bin all events
        weights_tot = self.weights * sigmav
        
        result = self._make_spec_hist(P, m, bins, bin_width, weights_tot, normalize)
        
        return result

    def _get_product_masses(self, which_product):
        """Return the masses to be used in the calls to relscatt."""
        if which_product == 1:
            m = self.m1
            mr = self.m2
            mr2 = self.m3
        elif which_product == 2:
            m = self.m2
            mr = self.m1
            mr2 = self.m3
        elif which_product == 3:
            if self.m3 is not None:
                m = self.m3
                mr = self.m1
                mr2 = self.m2
            else:
                raise ValueError('Invalid product choice (there are only two reaction products)')
        else:
            raise ValueError('Invalid product choice')

        return m, mr, mr2

    def _make_spec_hist(self, P, m, bins, bin_width, weights, normalize):

        E = P[0] - m           # kinetic energy (keV)
        valid = ~np.isnan(E)   # do not try to bin forbidden events
        E = E[valid]
        P = P[:,valid]
        weights = weights[valid]

        # Make attributes with product momenta and kinetic energies,
        # for possible later use
        self.P_prod = P
        self.E_prod = E

        # How to bin events
        if bins is None:
    
            return_bins = True

            if type(bin_width) == list:
                # Set up for computing 2D spectrum with default energy and pitch range
                bins_E = np.arange(E.min()-5*bin_width[0], E.max()+6*bin_width[0], bin_width[0])
                bins_A = np.arange(-1, 1+bin_width[1], bin_width[1])

                bins = [bins_E, bins_A]

            else:
                # Set up for computing energy spectrum with default bins
                bins = np.arange(E.min()-5*bin_width, E.max()+6*bin_width, bin_width)

        else:
            return_bins = False

        # Bin in 1D or 2D
        if type(bins) == list:
            # Bin in both energy and emission direction
            p = vec.normalize(P[1:])          # normalized emission directions
            A = vec.dot(p, self.ref_dir)      # cosine of emission angles w.r.t. the reference direction

            spec = np.histogram2d(E, A, bins=bins, weights=weights)[0]
        else:
            spec = np.histogram(E, bins=bins, weights=weights)[0]

        # Possibly normalize to unit sum
        if normalize:
            spec = spec / spec.sum()

        # Return results
        if return_bins:

            if spec.ndim == 2:
                # Return spectrum and bin edges of the 2D histogram
                return spec, bins[0], bins[1]

            else:
                # Return spectrum and bin centers of the 1D histogram
                E_axis = 0.5*(bins[1:] + bins[:-1])
                return spec, E_axis

        else:
            # Return only the spectrum, since the bins were supplied as input
            return spec
        




"""Classes for describing scattering reaction between alphas and d or t ions."""

import os

import numpy as np
from scipy.interpolate import RectBivariateSpline

from dress.config import cross_section_dir
from dress.reactions.particle import Particle
from dress.reactions.reaction import Reaction


def load_tab_cross_section(file_path):
    """Load tabulated differential cross section for alpha scattering."""

    f = open(file_path)

    # Dimensions of the data
    dims = tuple(int(n) for n in f.readline().split('\t'))

    cos_theta = np.zeros(dims[1])
    sigma_diff = np.zeros(dims)

    # Read energy axis
    E = np.array([float(e) for e in f.readline().strip().split('\t')])
    
    # Read cross section table
    for i,line in enumerate(f):
        vals = [float(val) for val in line.split('\t')]
        cos_theta[i] = vals[0]
        sigma_diff[:,i] = vals[1:]

    # Convert to the correct units
    E = E*1e3            # keV
    sigma_diff = sigma_diff*1e-31   # m**2 sr**-1

    # Calculate total cross section
    sigma_tot = 2*np.pi*np.trapz(sigma_diff, x=cos_theta, axis=1)

    return E, cos_theta, sigma_diff, sigma_tot


class AlphaScattering(Reaction):
    """A class representing alpha scattering on D or T."""

    __slots__ = ('sigma_tab', 'sigma_diff_tab', 'sigma_diff_interp', 'cos_theta_max')

    def __init__(self, d_or_t, cross_section_file):
        
        super().__init__(d_or_t, 'alpha', d_or_t, 'alpha', None)

        # Load tabulated differential cross section 
        # (same as the one used in ControlRoom)
        self.cx_path = os.path.join(cross_section_dir, cross_section_file)
        E, cos_theta, sigma_diff, sigma_tot = load_tab_cross_section(self.cx_path)

        sigma_tab = {}
        sigma_tab['E'] = E
        sigma_tab['cos_theta'] = cos_theta
        sigma_tab['sigma_diff'] = sigma_diff
        sigma_tab['sigma_tot'] = sigma_tot

        self.sigma_tab = sigma_tab
        self.sigma_diff_interp = RectBivariateSpline(E, cos_theta, sigma_diff, kx=1, ky=1)

        # Minimum scattering angle
        self.cos_theta_max = 1.0


    def _calc_sigma_tot(self, E):
        return np.interp(E, self.sigma_tab['E'], self.sigma_tab['sigma_tot'])

    def _calc_sigma_diff(self, E, cos_theta):
        xs = self.sigma_diff_interp(E, cos_theta, grid=False)
        xs[cos_theta>self.cos_theta_max] = 0.0
        return xs


class DAlphaScattering(AlphaScattering):
    """A class representing D + alpha -> D + alpha scattering."""

    def __init__(self):
        super().__init__('d', 'alphad-cross-section.txt')


class TAlphaScattering(AlphaScattering):
    """A class representing T + alpha -> T + alpha scattering."""

    def __init__(self):
        super().__init__('t', 'alphat-cross-section.txt')



"""The `tokamak` module is part of the `dress` package and contains 
tools for facilitating reaction product spectra calculations in tokamak
geometry, i.e. a toroidal system with toroidal symmetry, where quanteties
are typically a function either of (R,Z) or of a flux surface label ρ."""

import dress.tokamak.tokadist
import dress.tokamak.tokapart
import dress.tokamak.utils




"""Tools for calculating spatially resolved or volume integrated 
spectra with the `dress` framework."""

import numpy as np

import dress.vec_ops as vec

class VolumeElements:
    """A class representing volume elements from which we want to 
    compute reacion product spectra.

    Parameters
    ----------
    dV : array
        Volumes of each volume element (m**3)

    pos : tuple of arrays
        Spatial coordinates of each volume element, e.g. (R,Z) or (X,Y,Z).
        Not necessary for the spectrum calculations to work but could be useful
        for the user when plotting etc.

    ems_dir : array of shape (N,3) or None
        The emission direction along which to evaluate the spectrum. 
        `None` means to sample emission directions randomly in 4*pi.

    ref_dir : array of shape (N,3)
        If calculating spectra in 2D (i.e. resolved in both energy and emission direction), 
        the emission angle is given relative to this direction.

    solid_angle : array of shape (N,)
        Solid angle (in steradians) in which particles are emitted, for each volume element.

    dist_a, dist_b : instances of dress.dists.VelocityDistribution
        Distribution of the first and second reactants, respectively. Needs to be spatially
        resolved on the same grid as the volume elements, i.e. sampling the dists with the 
        keywords `index=i` should return a sample from volume element `i`."""

    def __init__(self, dV):
        """Create volume elements with given volumes `dV` (m**3)."""

        self.dV = np.atleast_1d(dV)
        self.nvols = len(dV)
        
        # Default attributes
        self.pos = None
        self.ems_dir = None
        self.ref_dir = [0,1,0]
        self.solid_angle = 4*np.pi
        self.dist_a = None
        self.dist_b = None

    @property
    def ems_dir(self):
        return self._ems_dir

    @ems_dir.setter
    def ems_dir(self, u):      

        if u is None or np.all(u == None):
            # Isotropic emission from all volume elements
            self._ems_dir = np.repeat(None, self.nvols)

        else:
            self._ems_dir = self._prep_vector_attribute(u)

    @property
    def ref_dir(self):
        return self._ref_dir

    @ref_dir.setter
    def ref_dir(self, u):
        self._ref_dir = self._prep_vector_attribute(u)
        
    def _prep_vector_attribute(self, v):
        """Put vector attribute (such as `ems_dir` and `ref_dir`) into the appropriate formats."""
        v = vec.make_vector(v)
            
        if v.shape[1] == 1:
            # Same vector for all volume elements
            v = vec.repeat(v, self.nvols)

        if v.shape[1] != self.nvols:
            raise ValueError('Number of vectors and volume elements do not match')

        # Transpose the array so that v[i] is the vector for element i
        v = v.T

        return v

    def _prep_scalar_attribute(self, x):
        """Put scalar attribute (such as `solid_angle`) into the appropriate format."""

        x = np.atleast_1d(x)
        
        if len(x) == 1:
            # Same value for all volume elements
            x = np.repeat(x, self.nvols)

        if len(x) != self.nvols:
            raise ValueError('Number values and volume elements do not match')

        return x

    @property
    def solid_angle(self):
        return self._solid_angle

    @solid_angle.setter
    def solid_angle(self, omega):
        self._solid_angle = self._prep_scalar_attribute(omega)


    @property
    def dist_a(self):
        return self._dist_a

    @dist_a.setter
    def dist_a(self, dist):
        
        if dist is not None and dist.n_spatial != self.nvols:
            raise ValueError('Wrong number of spatial positions for `dist_a`')

        self._dist_a = dist

    @property
    def dist_b(self):
        return self._dist_b

    @dist_b.setter
    def dist_b(self, dist):
        
        if dist is not None and dist.n_spatial != self.nvols:
            raise ValueError('Wrong number of spatial positions for `dist_b')

        self._dist_b = dist
    

if __name__ == '__main__':
    
    import matplotlib.pyplot as plt

    import dress
    import dists
    import vec_ops as vec

    # Create spectrum calculator
    dt = dress.reactions.DTNHe4Reaction()
    scalc = dress.SpectrumCalculator(dt)

    # Create a couple of volume elements
    nvols = 5
    dV = 1e-6*np.ones(nvols)
    
    vols = VolumeElements(dV)
    vols.dist_a = dists.MonoEnergeticDistribution(200.0*np.ones(nvols), scalc.reactant_a, 1e19*np.ones(nvols), pitch_range=[-0.1, 0.1])
    vols.dist_b = dists.MaxwellianDistribution(10*np.ones(nvols), scalc.reactant_b, 1e19*mp.ones(nvols))
    #vol.dist_b.v_collective = [0,1e6,0]

    vols.ems_dir = vec.repeat([0,0,1], nvols)
    vols.ref_dir = vec.repeat([0,1,0], nvols)

    # Calculate neutron spectrum
    E_bins = np.arange(11e3, 18e3, 50.0)
    A_bins = np.linspace(-1,1,25)
    #bins = [E_bins, A_bins]
    bins = E_bins
    spec = calc_vols(vols, scalc, bins, integrate=True)



import os

import numpy as np

from dress.reactions.masses import mp, m9Be
from dress.config import cross_section_dir
from dress.reactions.particle import Particle
from dress.reactions.reaction import Reaction


class PBe9DBe8Reaction(Reaction):
    """Class representing the p + Be9 -> d + 8Be fusion reaction."""

    __slots__ = ('tab_cross_section')

    def __init__(self):

        super().__init__('p', 'be9', 'd', 'be8', None)

        # Load tabulated cross section (from ENDF)
        cx_path = os.path.join(cross_section_dir,'Be9(p,d)Be8-cross-section_jsi.txt')
        tab_cross_section = np.loadtxt(cx_path, usecols=[0,1], comments='#')
        
        tab_cross_section[:,0] /= 1000.0                   # convert from eV to keV
        tab_cross_section[:,0] *= m9Be / (mp + m9Be)       # convert from LAB to CMS
        tab_cross_section[:,1] *= 1e-28                    # convert from barn to m**2

        self.tab_cross_section = tab_cross_section

    
    def _calc_sigma_tot(self, E):
        sigma =  np.interp(E, self.tab_cross_section[:,0], self.tab_cross_section[:,1],
                           left=0.0, right=0.0)

        return sigma

    def _calc_sigma_diff(self, E, costheta):
        E = np.atleast_1d(E)
        costheta = np.atleast_1d(costheta)

        # Assume isotropic cross section
        sigma = self.calc_sigma_tot(E)/(4*np.pi)    # m**2/sr

        return sigma



from dress.reactions.particle import Particle


class Reaction:
    """Base class for representing nuclear reactions with two or three particles in the final state."""

    __slots__ = ('name', 'a', 'b', 'p1', 'p2', 'p3')

    def __init__(self, a, b, p1, p2, p3):
        
        self.name = f'{a}-{b}'

        # Reactants
        self.a = Particle(a)
        self.b = Particle(b)

        # Products
        self.p1 = Particle(p1)
        self.p2 = Particle(p2)
        if p3 is None:
            self.p3 = None
        else:
            self.p3 = Particle(p3)

    def __repr__(self):
        return f'Reaction: {self.formula}'

    @property
    def formula(self):
        """A string representation of the reaction formula, like 'a + b -> 1 + 2'"""
        
        formula_string = f'{self.a.name} + {self.b.name} -> {self.p1.name} + {self.p2.name}'
        
        if self.p3 is not None:
            formula_string += f' + {self.p3.name}'            

        return formula_string

    @property
    def Q(self):
        m_reactants = self.a.m + self.b.m
        m_products = self.p1.m + self.p2.m

        if self.p3 is not None:
            m_products += self.p3.m

        return m_reactants - m_products

    def calc_sigma_tot(self, E):
        """Evaluate total cross section.

        Arguments
        ----------
        E : array-like
            Reactant energies in the center-of-momentum frame (keV).

        Returns
        -------
        array
            The cross section (in m**2) for each energy value.
        """
        return self._calc_sigma_tot(E)

    def calc_sigma_diff(self, E, costheta):
        """Evaluate angular differential cross section in the center-of-momentum frame.

        Arguments
        ---------
        E : array-like
            Reactant energies in the center of mass frame (keV).
        costheta : array-like
            Cosine of the emission angle of the neutron 
            (with respect to the reactant relative velocity)

        Returns
        -------
        array
            The cross section (in m**2 sr**-1) for each energy value."""
        return self._calc_sigma_diff(E, costheta)

        
    # Methods to be overloaded by sub-classes
    def _calc_sigma_tot(self,E):
        pass

    def _calc_sigma_tot(self,E, costheta):
        pass



from dress.reactions.d_d_n_he3_reaction import DDNHe3Reaction
from dress.reactions.d_t_n_he4_reaction import DTNHe4Reaction
from dress.reactions.d_he3_p_he4_reaction import DHe3PHe4Reaction
from dress.reactions.t_t_2n_he4_reaction import TT2NHe4Reaction
from dress.reactions.alpha_be9_n_c12_reaction import AlphaBe9NC12Reaction
from dress.reactions.alpha_be9_n_c12e1_reaction import AlphaBe9NC12E1Reaction
from dress.reactions.alpha_scat import DAlphaScattering, TAlphaScattering
from dress.reactions.p_be9_n_b9_reaction import PBe9NB9Reaction
from dress.reactions.p_be9_d_be8_reaction import PBe9DBe8Reaction
from dress.reactions.p_be9_alpha_li6_reaction import PBe9AlphaLi6Reaction
from dress.reactions.d_be9_n_b10_reaction import DBe9NB10Reaction



import numpy as np

from dress.reactions.particle import Particle
from dress.reactions.reaction import Reaction


class DTNHe4Reaction(Reaction):
    """Class representing the d + t -> n + he4 fusion reaction."""

    __slots__ = ()

    def __init__(self, reactant_a='d', reactant_b='t'):

        super().__init__('d', 't', 'n', '4he', None)

    
    def _calc_sigma_tot(self, E):        
        E = np.atleast_1d(E).astype('d')
        sigma = np.zeros_like(E)
        
        # Bosch-Hale parameterization of the T(d,n)4He reaction 
        B_G  = 34.3827
        A1_l = 6.927e4
        A2_l = 7.454e8
        A3_l = 2.05e6
        A4_l = 5.2002e4
        A5_l = 0.0
        B1_l = 6.38e1
        B2_l = -9.95e-1
        B3_l = 6.981e-5
        B4_l = 1.728e-4
        A1_h = -1.4714e6
        A2_h = 0.0
        A3_h = 0.0
        A4_h = 0.0
        A5_h = 0.0
        B1_h = -8.4127e-3
        B2_h = 4.7983e-6
        B3_h = -1.0748e-9
        B4_h = 8.5184e-14

        I_l = E<=550.0 
        I_h = E>550.0
        El = E[I_l]
        Eh = E[I_h]
        
        # Calculate the S factor
        S = np.zeros(np.shape(E))
        S[I_l] = (A1_l + El*(A2_l + El*(A3_l + El*(A4_l + El*A5_l))))/\
                (1.0 + El*(B1_l + El*(B2_l + El*(B3_l + El*B4_l))))
        S[I_h] = (A1_h + Eh*(A2_h + Eh*(A3_h + Eh*(A4_h + Eh*A5_h))))/\
                (1.0 + Eh*(B1_h + Eh*(B2_h + Eh*(B3_h + Eh*B4_h))))
        
        # Calculate the cross section
        nonzero = E > 0
        sigma[nonzero] = S[nonzero]/(E[nonzero]*np.exp(B_G/np.sqrt(E[nonzero])))     # mb
        
        return sigma * 1e-31     # m**2

    def _calc_sigma_diff(self, E, costheta):
        E = np.atleast_1d(E)
        costheta = np.atleast_1d(costheta)

        # Assume isotropic cross section
        sigma = self.calc_sigma_tot(E)/(4*np.pi)    # m**2/sr

        return sigma



"""A module for holding and processing ion distribution data for
tokamaks, where the distribution is typically given as functions of 
two spatial coordinates (R,Z)."""


import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import interp1d


class TokaDistData:
    """A class for holding distribution data.

    The distribution data is held in an array `F`, organized such that
    `F[i]` gives the distribution data at spatial point with index `i`.

    The distribution data at a given spatial point. Could be a scalar 
    (e.g. temperature) or an array representing e.g. f(v), f(E,pitch) etc."""

    def __init__(self, dist_data, density_data, axes):
        """Initialize a `TokaDistData` instance.

        Parameters
        -----------

        dist_data : array
            The distribution data. The length of the first dimension 
            should match the number of spatial points `NP`.

        density : array of shape (NP,)
            The density at each spatial point.

        axes : tuple of arrays
            The axis along each dimension (e.g. speed, pitch , energy,...).
            The spatial dimension is handeled separately (by the 
            `spatial_index_fun` argument) and is NOT included here."""


        self.F = dist_data
        self.density = density_data
        self.X = axes


    @property
    def F(self):
        return self._F

    @F.setter
    def F(self, dist_data):
        dist_data = np.atleast_2d(dist_data)
        
        # Add one last point with only zeros
        dist_shape = list(dist_data.shape)
        dist_shape[0] = dist_shape[0] + 1

        F = np.zeros(dist_shape)
        F[:-1] = dist_data

        # Set attribute
        self._F = F

    @property
    def density(self):
        return self._density

    @density.setter
    def density(self, n):
        n = np.atleast_1d(n)
        
        # Add one last point with zero density
        density = np.zeros(len(n)+1)
        density[:-1] = n

        # Set attribute
        self._density = density

    def _get_spatial_index(self, R, Z):
        """Return the spatial index for the given (R,Z) values."""
        pass         # to be over-ridden by subclasses


    def map_dist(self, R, Z):
        """Return distributions and densities at the given (R,Z) points."""

        R, Z = np.atleast_1d(R, Z)
        i_spatial = self._get_spatial_index(R, Z)
        F = self.F[i_spatial]
        density = self.density[i_spatial]

        return F, density

        
# Subclasses for common special cases
# -----------------------------------

class RhoDistData(TokaDistData):
    """Class for holding distribution data where spatial variations 
    depend on a flux surface label `rho` only."""

    def __init__(self, dist_data, density_data, axes, rho_axis, flux_map):
        """Initialize dist which is a function of rho only.

        Parameters
        ----------

        rho_axis : array
            The rho axis used by the distribution.

        flux_map : dress.tokamak.utils.FluxSurfaceMap
            Mapping between (R,Z) and rho.""" 
        
        super().__init__(dist_data, density_data, axes)
        self.rho = rho_axis

        rho_indices = np.arange(len(rho_axis))
        ind_fun = interp1d(rho_axis, rho_indices, kind='nearest', 
                           bounds_error=False, fill_value=-1)

        self.ind_fun = ind_fun
        self.flux_map = flux_map


    @property
    def rho(self):
        return self._rho

    @rho.setter
    def rho(self, x):
        rho_last = x[-1]*1.1
        _rho = np.zeros(len(x)+1)
        _rho[:-1] = x
        _rho[-1] = rho_last
        self._rho = _rho


    def _get_spatial_index(self, R, Z):
        rho_RZ = self.flux_map.get_rho(R, Z)
        ind_RZ = self.ind_fun(rho_RZ)

        return ind_RZ.astype('int')



"""Module for relativistic scattering calculations."""

import numpy as np
from scipy.constants import c       # speed of light

from dress import relkin

try:
    import sys
    sys.path.append('/home/jeriks/Utils/jspec-res/')

    import jspecf
    CAN_DO_THREEBODY = True

except ModuleNotFoundError:
    CAN_DO_THREEBODY = False


class ReactantData:
    """Helper class for calculating and holding kinematic data about reactants."""

    def __init__(self, Pa, Pb, ma, mb):
        
        # Basic quanteties
        self.Pa = Pa
        self.Pb = Pb
        self.ma = ma
        self.mb = mb

        self.Ptot = Pa + Pb
        self.Mtot = np.sqrt(relkin.mult_four_vectors(self.Ptot, self.Ptot))      # total invariant mass

        self.Tcm = self.Mtot - ma - mb    # kinetic energy in the CMS
        
        # Gamma and beta factors for the CMS
        self.gamma_cm = self.Ptot[0] / self.Mtot
        self.beta_cm =  self.Ptot[1:] / self.Ptot[0]

        # Relative velocity
        self.vrel = calc_vrel(Pa, Pb)
        self.vrel_mag = np.linalg.norm(self.vrel, axis=0)


class ProductData:
    """Helper class for calculating and holding kinematic data about reaction products."""

    def __init__(self, P, beta_cm):

        # Lab frame
        self.P = P
        self.pmag = np.linalg.norm(P[1:], axis=0)

        # CMS
        self.Pcm = relkin.boost(P, beta_cm)
        self.pcm_mag = np.linalg.norm(self.Pcm[1:], axis=0)


def two_body_event(reactant_data, m1, m2, u1):
    """
    Calculate a possible 4-vector (along direction 'u1') of particle 1 in
    the two-body event

           a + b --> 1 + 2

    'Pa' and 'Pb' are the incoming 4-momenta (keV/c).
    'm1' and 'm2' are the masses of the products (keV/c**2).

    All vector input should be 2D, i.e. shape (4,N) for four-vectors
    and (3,N) for three-vectors, where N is either equal to 1 or to
    the number of events.
    """

    # Normalize emission directions
    u1 = u1 / np.linalg.norm(u1, axis=0)

    # Calculate magnitude of 3-momentum
    p1_mag = invariant_to_momentum(m2**2, reactant_data, m1, u1)

    # Energy
    E1 = np.sqrt(p1_mag**2 + m1**2)

    # 4-vector
    P1 = relkin.four_vector(E1, p1_mag*u1)

    return P1


def three_body_event(reactant_data, m1, m2, m3, u1):
    """
    Calculate a possible 4-vector (direction 'u1') of particle 1 in
    the three-body event

         a + b --> 1 + 2 + 3

    'pa' and 'pb' are the incoming 4-momenta (keV/c). 'm1', 'm2' and 'm3' are
    the masses of the products (keV/c**2).

    All vector input should be 2D, i.e. shape (4,N) for four-vectors
    and (3,N) for three-vectors, where N is either equal to 1 or to
    the number of events.
    """
    
    if not CAN_DO_THREEBODY:
        raise NotImplementedError('Three-body calculations not available')

    # Normalize emission directions
    u1 = u1 / np.linalg.norm(u1, axis=0)

    # Total invariant mass
    Mtot = reactant_data.Mtot

    # Sample available phase space (Dalitz plot)
    m23_sq = np.zeros_like(Mtot)
    for i,mtot in enumerate(Mtot):
        m12_sq, m23_sq[i] = jspecf.relevents.dalitzsample(mtot, m1, m2, m3) 

    # Calculate magnitude of 3-momentum
    p1_mag = invariant_to_momentum(m23_sq, reactant_data, m1, u1)

    # Energy
    E1 = np.sqrt(p1_mag**2 + m1**2)

    # 4-vector
    P1 = relkin.four_vector(E1, p1_mag*u1)

    return P1


def invariant_to_momentum(inv, reactant_data, m1, u1):
    """Solution to the four-momentum conservation equation."""

    Ptot = reactant_data.Ptot
    Mtot = reactant_data.Mtot

    Cm  = Mtot**2 + m1**2 - inv
    p0u = np.sum(Ptot[1:] * u1, axis=0)    # projection of CM momentum on emission direction
    E0  = Ptot[0]

    first_term = Cm * p0u

    sq_term1 = Cm**2 * E0**2
    sq_term2 = 4*m1**2 * E0**2 * (E0**2 - p0u**2)
    second_term = np.sqrt(sq_term1 - sq_term2)

    p1_plus = (first_term + second_term) / (2 * (E0**2 - p0u**2))   # magnitude of three-momentum (1st solution)
    p1_minus = (first_term - second_term) / (2 * (E0**2 - p0u**2))   # magnitude of three-momentum (2nd solution)

    # Check which solution (if any) that is valid
    valid_plus = p1_plus >= 0.0
    valid_minus = p1_minus >= 0.0

    p1 = np.repeat(np.nan, len(p1_plus))
    p1[valid_minus] = p1_minus[valid_minus]
    p1[valid_plus] = p1_plus[valid_plus]

    has_two_solutions = valid_plus & valid_minus
    if np.any(has_two_solutions):
        n = np.sum(has_two_solutions)
        ntot = len(has_two_solutions)
        print(f'NOTE: two kinematically allowed solutions found in {n}/{ntot} cases. Using the one with highest energy.')

    return p1


def calc_vrel(Pa, Pb):
    """Calculate relative velocity between particles with four-momenta ´Pa´ and ´Pb´."""
    
    beta_b = Pb[1:] / Pb[0]
    Pa_b = relkin.boost(Pa, beta_b)   # boost particle a into the rest frame of particle b
    vrel = Pa_b[1:] / Pa_b[0]
    
    return vrel


def calc_jacobian_cms_lab(reactant_data, product_data):
    """
    Jacobian for transforming between CMS and LAB emission solid angles,
    as given by Eq. (III.4.16) in Byckling & Kajantie.
    """

    P = product_data.P
    pmag = product_data.pmag
    pcm_mag = product_data.pcm_mag

    beta_cm = reactant_data.beta_cm
    gamma_cm = reactant_data.gamma_cm

    u_dot_beta = np.sum(P[1:]*beta_cm, axis=0) / pmag
    jacobian = pmag**2 / (gamma_cm * pcm_mag * (pmag - P[0]*u_dot_beta))

    return jacobian


def calc_costheta_cm(reactant_data, product_data):
    """Calculate cosine of emission angle of a given reaction product.

    The emission angle is the angle between the emission direction and the
    reactant relative velocity, in the CMS.
    """

    vrel = reactant_data.vrel
    vrel_mag = reactant_data.vrel_mag
    
    nonzero = vrel_mag > 0
    u_rel = np.zeros_like(vrel)
    u_rel[:,nonzero] = vrel[:,nonzero] / vrel_mag[nonzero]   # unit vector along relative motion of reactants

    P1_cm = product_data.Pcm
    p1cm_mag = product_data.pcm_mag
    costheta = np.sum(P1_cm[1:]*u_rel, axis=0) / p1cm_mag

    return costheta
    

def get_reactivity(reactant_data, P1, reaction):
    """
    Calculate the reactivity (m**3/sr/s) for a reaction with
    given reactant data and the product of interest with four-momentum ´P1´.

    All vector input should be 2D, i.e. shape (4,N) for four-vectors
    and (3,N) for three-vectors, where N is either equal to 1 or to
    the number of events.
    """

    # Eval P1 in the CMS
    P1_data = ProductData(P1, reactant_data.beta_cm)

    # Jacobian for transforming between CMS and LAB
    jacobian = calc_jacobian_cms_lab(reactant_data, P1_data)

    # Angular differential cross section in the CMS
    costheta = calc_costheta_cm(reactant_data, P1_data)
    sigma = reaction.calc_sigma_diff(reactant_data.Tcm, costheta)

    # Reactivity (m**3/sr/s)
    vrel_mag = reactant_data.vrel_mag * c      # m/s
    sigmav = sigma * vrel_mag * jacobian

    return sigmav



"""Routines for sampling different types of distributions."""

import numpy as np
import scipy.integrate as integrate


def sample_mono(val, n_samples):
    """ Sample delta function. """

    n_samples = int(n_samples)
    s = val * np.ones(n_samples)

    return s

def sample_uniform(val_range, n_samples):
    """ Sample values from uniform distribution. """
    
    n_samples = int(n_samples)
    s = np.random.uniform(val_range[0], val_range[1], size=n_samples)

    return s

def sample_acc_rej(f, lims, fmax, n_samples, quiet=True):
    """ Randomly sample from a distribution f, using the acceptance-rejection method.
    'f' should be callable, taking the dependent variables input. The dimensionality of f
    is inferred from the tuple 'lims', which should contain the ranges to sample from, i.e.

         lims = ([x0_min,x1_min, ...], [x0_max,x1_max, ...])

    'fmax' is the maximum number that f can take. """

    n_samples = int(n_samples)

    # Sample points xp and fp in the hypercube bounded by 'lims' and 'fmax', and keep only
    # the points that fall below f(*xp). Loop until we have collected enough samples.
    ndims = len(lims[0])

    s = np.zeros((n_samples, ndims))
    n_to_sample = n_samples
    n_sampled = 0

    while n_to_sample > 0:

        # Sample points
        xp = np.random.uniform(low=lims[0], high=lims[1], size=(n_to_sample,ndims))
        fp = np.random.uniform(low=0, high=fmax, size=n_to_sample)

        # Accept/reject
        x_acc = xp[fp < f(*xp.T)]

        # Save accepted points
        i0 = n_sampled

        n_sampled += len(x_acc)
        i1 = n_sampled

        s[i0:i1] = x_acc

        # How many points to sample next time
        n_to_sample -= len(x_acc)

        if not quiet:
            print('{:.2f}%'.format(n_sampled/n_samples*100))

    return s

def sample_sphere(n_samples, pitch_range=(-1, 1)):
    """Sample points that are uniformly distributed over the surface
    of a unit sphere. Output points are given in a rectangular coordinate 
    system and is given as an array with shape (3, n_samples). """

    n_samples = int(n_samples)

    # Uniformly sample cosine of polar angle
    u = np.random.uniform(*pitch_range, size=n_samples)

    # Uniformly sample the azimuthal angle
    theta = np.random.uniform(0, 2*np.pi, size=n_samples)

    # Coordinates in a rectangular system
    x = np.sqrt(1-u**2) * np.cos(theta)
    y = np.sqrt(1-u**2) * np.sin(theta)
    z = u

    # Return array
    v = np.vstack((x,y,z))

    return v

def sample_mcmc(F, x0, dx, n_samples):
    """
    Sample F(x) by means of Markov Chain Monte Carlo sampling. 
    'F' should be callable, accepting a vector 'x' as input. 
    'x0' is the starting point for the Markov chain.
    'dx' is a suitable step size.
    """

    n_samples = int(n_samples)

    # Parameters for controlling how to step to new points
    mean_step = np.zeros_like(x0)
    cov_step = dx * np.eye(len(x0))
    
    # Array to store the resulting Markov chain
    chain = np.zeros((n_samples,len(x0)))

    for i in range(n_samples):
        
        # Evaluate function at the present point
        f0 = F(x0)
        
        # Step to next point
        x1 = x0 + np.random.multivariate_normal(mean_step, cov_step)

        # Evaluate function at the new point
        f1 = F(x1)

        # Calculate the Markov ratio
        r = f1/f0

        # Accept new point unconditionally if r >= 1,
        # but if r < 1, accept it with probability r
        if r < 1:
            u = np.random.rand()

            if u > r:
                x1 = x0

        # Store new point and roll values for next step
        chain[i] = x1
        x0 = x1

    return chain

def sample_tab(dist, *axes, n_samples=1e6, dx=None, var_type='continuous'):
    """Sample a tabulated probability density function (pdf).

    Parameters
    ----------
    dist : array-like of shape (N0, N1, N2, ...)
        The tabulated probability density function. Does not have to be normalized.
    axes : 1D-arrays
        The coordinate axes (bin centers). The number of axes should match the number of 
        dimensions of `dist`. The length of each axis should match the number of elements 
        along the corresponding dimension of `dist`, i.e. len(axes[0]) = N0, 
        len(axes[1]) = N1 etc.
    n_samples : int
        Number of samples to draw.
    dx : list of 1D arrays or None
        The widths of the bins represented by the axes. If `None` (default) an attempt is
        made to reconstruct the bin widths from the axes arrays.
    var_type : str
        Can be either 'continuous', in which case the sampled values are distributed uniformly
        within each bin, of 'discrete', in which case only the exact tabulated values are sampled.

    Returns
    -------
    sample : array of shape (dist.ndim, n_samples)
        Coordinates of the random samples."""

    n_samples = int(n_samples)
    dist = np.atleast_1d(dist)
    axes = [np.atleast_1d(x) for x in axes]
    axes_mgrid = np.meshgrid(*axes, indexing='ij')

    if dx is None:
        dx = [None]*dist.ndim

    for i in range(len(dx)):
        if dx[i] is not None:
            dx[i] = np.atleast_1d(dx[i])

    # Input checks
    if len(axes) != dist.ndim:
        raise ValueError(f'Number of axes must match the number of dimenstion in `dist`')

    if len(dx) != dist.ndim:
        raise ValueError(f'Number of axes must match the number of dimenstion in `dist`')

    for i,x in enumerate(axes):
        if x.shape != (dist.shape[i],):
            raise ValueError(f'Axis {i} with shape {x.shape} incompatible with dist with shape {dist.shape}')

        if dx[i] is not None:
            if dx[i].shape != (dist.shape[i],):
                raise ValueError(f'Bin widths for axis {i} with shape {dx[i].shape} incompatible with dist with shape {dist.shape}')
    
    # Determine the volume elements
    for i,dx_ in enumerate(dx):
        if dx_ is None:
            dx_ = _reconstruct_bin_widths(axes[i])
            dx[i] = dx_
        
    dx = np.meshgrid(*dx, indexing='ij')
    dv = 1.0
    for dx_ in dx:
        dv = dv*dx_

    # Flatten the distribution to a 1D pdf    
    pdf_1d = dist.flatten()
    x_1d = [x.flatten() for x in axes_mgrid]
    dx_1d = [dx_.flatten() for dx_ in dx]
    dv_1d = dv.flatten()

    # Multiply pdf and dv to get a probability distribution (rather than a density)
    P = pdf_1d * dv_1d
    P = P/P.sum()     # normalize to unity
    
    # Sample from the probability distribution
    i_sample = np.random.choice(len(P), p=P, size=n_samples)
    sample = np.array([x[i_sample] for x in x_1d])

    if var_type == 'continuous':
        # Redistribute each sample uniformly within its respective bin
        jitter = (-0.5 + np.random.random_sample(sample.shape))     # random variable from [-0.5,0.5]
        dx_sample = np.array([dx[i_sample] for dx in dx_1d])        # bin width for each sample
        sample = sample + jitter*dx_sample
        
    elif var_type != 'discrete':
        raise ValueError('`var_type` must be either "continuous" or "discrete"')
        
    
    sample = sample.squeeze()        # remove singleton dimension in case of 1D dist

    return sample
    
def _reconstruct_bin_widths(bin_centers):
    """Determine the bin widths for a sequence of bin_centers.

    For bins with variable bin spacing this is a bit ambigous;
    here we make the assumption that the bin widths are everywhere
    equal to the distance between successive bin centers, and that
    the last bin width is equal to the second last bin width."""

    w = np.zeros_like(bin_centers)
    w[:-1] = np.diff(bin_centers)
    w[-1] = w[-2]

    return w



"""Useful methods for performing `dress` calculations."""

import numpy as np

from dress import dists
from dress import volspec
from dress.reactions.particle import Particle

def make_dist(dist_type, particle_name, NP, density, **kwargs):
    """Create distribution from the given input data.

    Parameters
    ----------
    dist_type : str
        The type of dist. Should be one of ('maxwellian', 'energy', 'speed', 
        'energy-pitch', 'vpar-vperp')

    particle_name : str
        Name of the particle (e.g. 'd', 't, 'he4' etc). See `dress.reactions.particle.py`
        for all valid particle names.

    NP : int
        Number of spatial points covered by the distribution

    density : array of shape (NP,)
        Density (particles/m**3) at each of the NP spatial points

    All distributions require the above input. Additionally, depending on the requested distribution type,
    some of the following keyword arguments should be supplied:

    General keyword arguments
    -------------------------    
    v_collective : array-like of shape (NP,3)
        Collective velocity (m/s) at each of the NP spatial points. 
        Default is no collective velocity.

    ref_dir : array-like with shape (NP,3)
        Reference direction at each of the NP spatial points.
        Pitch values of the particles are given relative to this direction.
        Does not have to be normalized to unity.
        Default is [0,1,0].


    Keyword arguments for `dist_type = 'maxwellian'`
    ---------------------------------------------
    temperature : array-like of shape (NP,)
        Temperature (keV) at each of the NP spatial points

    pitch_range : array-like of shape (2,)
        Range of (uniformly distributed) pitch values.
        Default is isotropically distributed velocities, i.e. pitch_range = [-1,1].


    Keyword arguments for `dist_type = 'energy'`
    --------------------------------------------
    energy_axis : array-like of shape (NE)
        Energy axis of the tabulated distribution

    distvals : array of shape (NP,NE)
        Tabulated energy distribution at each of the NP spatial points.
    
    pitch_range : array-like of shape (2,)
        Range of (uniformly distributed) pitch values.
        Default is isotropically distributed velocities, i.e. pitch_range = [-1,1].

    
    Keyword arguments for `dist_type = 'speed'`
    --------------------------------------------
    speed_axis : array-like of shape (Nv)
        Speed axis of the tabulated distribution

    distvals : array of shape (NP,Nv)
        Tabulated speed distribution at each of the NP spatial points.
    
    pitch_range : array-like of shape (2,)
        Range of (uniformly distributed) pitch values.
        Default is isotropically distributed velocities, i.e. pitch_range = [-1,1].


    Keyword arguments for `dist_type = 'energy-pitch'`
    --------------------------------------------------
    energy_axis : array-like of shape (NE,)
        Energy axis of the tabulated distribution

    pitch_axis : array-like of shape (Npitch,)
        Pitch axis of the tabulated distribution

    distvals : array of shape (NP,NE,Npitch)
        Tabulated energy-pitch distribution at each of the NP spatial points.


    Keyword arguments for `dist_type = 'vpar-vperp'`
    --------------------------------------------------
    vpar_axis : array-like of shape (Nvpar,)
        v_parallel axis of the tabulated distribution

    vperp_axis : array-like of shape (Nvperp,)
        v_perpendicular axis of the tabulated distribution

    distvals : array of shape (NP,Nvpar,Nvperp)
        Tabulated vpar-vperp distribution at each of the NP spatial points.


    Returns
    -------
    dist : instance of dress.dists.VelocityDistribution
        The resulting distribution in `dress` format."""

    
    # Massage the input a bit
    dist_type = dist_type.lower()
    particle = Particle(particle_name)
    NP = int(NP)
    density = np.atleast_1d(density)
    if not density.shape == (NP,): raise ValueError(f'Wrong shape of `density`, should be ({NP},)')

    # Get keyword arguments
    default_v_collective = np.zeros((NP,3))
    v_collective = _get_dist_kwarg(kwargs, 'v_collective', (NP,3), default_v_collective)
    v_collective = v_collective.T     # make correct shape for dress

    default_ref_dir = np.repeat(np.atleast_2d([0,1,0]), NP, axis=0)
    ref_dir = _get_dist_kwarg(kwargs, 'ref_dir', (NP,3), default_ref_dir)
    ref_dir = ref_dir.T     # make correct shape for dress

    default_pitch_range = [-1,1]
    pitch_range = _get_dist_kwarg(kwargs, 'pitch_range', (2,), default_pitch_range)

    temperature = _get_dist_kwarg(kwargs, 'temperature', (NP,), None)
    energy_axis = _get_dist_kwarg(kwargs, 'energy_axis', None, None)
    speed_axis = _get_dist_kwarg(kwargs, 'speed_axis', None, None)
    pitch_axis = _get_dist_kwarg(kwargs, 'pitch_axis', None, None)
    vpar_axis = _get_dist_kwarg(kwargs, 'vpar_axis', None, None)
    vperp_axis = _get_dist_kwarg(kwargs, 'vpar_axis', None, None)

    # Create the requested distribution
    if dist_type == 'maxwellian':
        dist = dists.MaxwellianDistribution(temperature, particle, density=density, v_collective=v_collective,
                                            pitch_range=pitch_range, ref_dir=ref_dir)

    elif dist_type == 'energy':
        if energy_axis is None: raise ValueError('Must provide `energy_axis` keyword')
        NE = len(energy_axis)

        distvals = _get_dist_kwarg(kwargs, 'distvals', (NP,NE), None)
        if distvals is None: raise ValueError('Must provide `distvals` keyword')
        
        dist = dists.TabulatedEnergyDistribution(energy_axis, distvals, particle, density=density,
                                                 pitch_range=pitch_range, ref_dir=ref_dir)

    elif dist_type == 'speed':
        if speed_axis is None: raise ValueError('Must provide `speed_axis` keyword')
        Nv = len(speed_axis)

        distvals = _get_dist_kwarg(kwargs, 'distvals', (NP,Nv), None)
        if distvals is None: raise ValueError('Must provide `distvals` keyword')
        
        dist = dists.TabulatedSpeedDistribution(speed_axis, distvals, particle, density=density,
                                                pitch_range=pitch_range, ref_dir=ref_dir)

    elif dist_type == 'energy-pitch':
        if energy_axis is None: raise ValueError('Must provide `energy_axis` keyword')
        NE = len(energy_axis)

        if pitch_axis is None: raise ValueError('Must provide `pitch_axis` keyword')
        Npitch = len(pitch_axis)

        distvals = _get_dist_kwarg(kwargs, 'distvals', (NP,NE,Npitch), None)
        if distvals is None: raise ValueError('Must provide `distvals` keyword')
        
        dist = dists.TabulatedEnergyPitchDistribution(energy_axis, pitch_axis, distvals, particle, 
                                                      density=density, ref_dir=ref_dir)

    elif dist_type == 'vpar-vperp':
        if vpar_axis is None: raise ValueError('Must provide `vpar_axis` keyword')
        Nvpar = len(vpar_axis)

        if vperp_axis is None: raise ValueError('Must provide `vperp_axis` keyword')
        Nvperp = len(vperp_axis)

        distvals = _get_dist_kwarg(kwargs, 'distvals', (NP,Nvpar,Nvperp), None)
        if distvals is None: raise ValueError('Must provide `distvals` keyword')
        
        dist = dists.TabulatedVparVperpDistribution(vpar_axis, vperp_axis, distvals, particle, 
                                                    density=density, ref_dir=ref_dir)

    else:
        raise ValueError(f'dist_type = {dist_type} is not a valid option')

    return dist


def _get_dist_kwarg(kwargs, kw, required_shape, default_val):
    var = kwargs.get(kw, default_val)
    
    if var is not None:
        var = np.atleast_1d(var)

        if (required_shape is not None) and (var.shape != required_shape):
            raise ValueError(f'Shape of {kw} is {var.shape}, but it should be {required_shape}')

    return var


def make_vols(dV, solid_angle, emission_dir=None, ref_dir=None, pos=None):
    """Make volume elements for dress.

    Make a collection of volume elements for which volume intergated or 
    spatially resolved spectrum calculations can be performed.

    Parameters
    ----------
    dV : array-like of shape (NP,)
        Element volumes (in m**3).

    solid_angle : array-lilke of shape (NP,)
        Solid angle that the reaction products of interest are emitted into
        (typically defined by the viewing geometry, collimation etc).

    emission_dir : array-like of shape (NP,3) or None (default)
        Reaction product emission direction for each volume element. Does not 
        have to be normalized to unit length. emission_dir = None (default) means 
        isotropic emission in 4*pi.

    ref_dir : array-like of shape (NP,3)
        Reference direction for the emitted reaction products. Only needs to be
        provided if the spectrum should be resolved with respect to the emission angle.

    pos : array or tuple of arrays
        Spatial coordinates of each volume element, e.g. (x, y, z), (R, Z) or similar.
        Not required for the dress calculations, but can be handy for plotting etc.
    
    Returns
    -------
    vols : dress.volspec.VolumeElement
        The resulting volume elements."""

    # Check the input
    dV = np.atleast_1d(dV)
    if dV.ndim != 1: 
        raise ValueError('`dV` should be a 1D array')

    NP = len(dV)

    solid_angle = np.atleast_1d(solid_angle)
    if solid_angle.shape != (NP,): 
        raise ValueError('`solid_angle` should have the same length as `dV`')
    
    if emission_dir is not None:
        emission_dir = np.atleast_2d(emission_dir)
        
        if emission_dir.shape != (NP,3):
            raise ValueError('Wrong shape of `emission_dir`')

        emission_dir = emission_dir.T    # transpose to get correct shape for dress

    if ref_dir is None:
        # User does not intend to resolve spectra with respect to 
        # emission direction -> ref_dir is arbitrary
        ref_dir = np.atleast_2d([0,1,0])
        ref_dir = np.repeat(ref_dir, NP, axis=0)
        ref_dir = ref_dir

    else:
        ref_dir = np.atleast_2d(ref_dir)
        
    if ref_dir.shape != (NP,3):
        raise ValueError('Wrong shape of `ref_dir`')

    ref_dir = ref_dir.T   # transpose to get correct shape for dress

    # Create volume elements
    vols = volspec.VolumeElements(dV)
    vols.solid_angle = solid_angle
    vols.ems_dir = emission_dir
    vols.ref_dir = ref_dir
    vols.pos = pos

    return vols


def calc_vols(vols, dist_a, dist_b, spec_calc, bins, integrate=True, quiet=True, **kwargs):
    """Calculate spectrum from a number of volume elements and corresponding reactant distributions.

    Parameters
    ----------
    vols : instance of dress.volspec.VolumeElements
        The volume elements to calculate the spectra from.

    dist_a, dist_b : instances of dress.dists.VelocityDistribution
        The reactant distributions to sample from.

    spec_calc : instance of dress.spec.SpectrumCalculator
        The spectrum calculator to apply to each volume element.

    bins : array or [array, array]
        Bin edges in energy and (optionally) in pitch of the emission direction. 

    integrate : bool
        Whether to calculate the volume integrated spectrum (units of particles/bin/s)
        or a spatially resolved spectrum (units of particles/bin/m**3/s).

    quiet : bool
        Whether to print a status message about the progress of the spectrum calculations.

    All additional keyword arguments are passed to `spectrum_calculator.__call__`.

    Returns
    -------
    spec : array
        If integrate = True this will be a 1D array representing the volume integrated 
        spectrum histogram. If integrate = False this will be a 2D array with N rows, 
        such that spec[i] is the spectrum from vols[i]."""

    vols.dist_a = dist_a
    vols.dist_b = dist_b

    spec = _get_empty_spec(vols, bins, integrate)

    for i in range(vols.nvols):

        if not quiet:
            print(f'Progress: {100*i/vols.nvols:.2f}%', end='\r')

        # Spectrum from current volume element (particles/bin/m**3/s)
        s = calc_single_vol(vols, i, spec_calc, bins=bins, **kwargs)

        if integrate:
            # Compute volume integrated spectrum (with units particles/bin/s)
            spec = spec + s*vols.dV[i]
        else:
            # Compute spatially resolved spectrum (with units particles/bin/m**3/s)
            spec[i] = s

    return spec


def calc_single_vol(vols, index, spec_calc, **kwargs):
    """Calculate spectrum from a given volume element.

    Parameters
    ----------
    vols : instance of dress.volspec.VolumeElements
        The volume elements to calculate the spectrum from.

    index : int
        The index of the volume element to calculate ths spectrm from.

    spec_calc : instance of dress.spec.SpectrumCalculator
        The spectrum calculator to apply to the volume element.

    All additional keyword arguments are passed to `spectrum_calculator.__call__`.

    Returns
    -------
    spec : array
        The calculated spectrum (units are particles/bin/m**3/s)."""
    
    if (spec_calc.reaction.a != vols.dist_a.particle or 
        spec_calc.reaction.b != vols.dist_b.particle):
        raise ValueError('Reactants and distribution species do not match')
    
    # Check if dist is nonzero
    na = vols.dist_a.density[index]
    nb = vols.dist_b.density[index]
    ΔΩ = vols.solid_angle[index]
    δab = get_delta(vols.dist_a, vols.dist_b)

    if 0.0 in [na, nb, ΔΩ]:
        return 0.0

    # Calculate spectrum along the requested emission direction
    spec_calc.u = vols.ems_dir[index]
    spec_calc.ref_dir = vols.ref_dir[index]

    n_samples = spec_calc.n_samples
    
    spec_calc.weights = na*nb*ΔΩ/(1 + δab) * np.ones(n_samples) / n_samples

    # Sample reactant distributions
    spec_calc.reactant_a.v = vols.dist_a.sample(spec_calc.n_samples, index=index)
    spec_calc.reactant_b.v = vols.dist_b.sample(spec_calc.n_samples, index=index)

    spec = spec_calc(**kwargs)

    return spec


def _get_empty_spec(vols, bins, integrate):
    """Create empty spectrum array of the appropriate size."""
    
    # Make `bins` into 1-element lists, if necessary
    if type(bins) is not list:
        bins = [bins]

    # Check if we should have spatial resolution or not
    if integrate:
        nvols = 1
    else:
        nvols = vols.nvols
    
    # Number of energy bins
    nE = len(bins[0]) - 1
    
    # Check if spectrum should be resolved in emission direction as well
    if len(bins) == 2:
        nA = len(bins[1]) - 1
    else:
        nA = 1
    
    # Create empty spectrum array of correct shape
    spec = np.zeros((nvols,nE,nA))
    
    # Remove unecessary dimensions
    spec = np.squeeze(spec)

    return spec


def get_delta(dist_a, dist_b):
    
    if dist_a == dist_b:
        delta = 1
    else:
        delta = 0

    return delta


if __name__ == '__main__':
    
    NP = 1
    density = np.repeat(1e19, NP)
    particle_name = 'd'

    # Make a Maxwellian distribution
    T = np.repeat(10, NP)
    maxwellian_dist = make_dist('maxwellian', particle_name, NP, density, temperature=T)

    # Make tabulated energy distribution
    Eaxis = np.linspace(0, 1000, 500)
    Edist = np.zeros_like(Eaxis)
    Edist[:250] = Eaxis[:250]
    Edist[250:] = Eaxis[-1] - Eaxis[250:]
    Edist = np.repeat(np.atleast_2d(Edist), NP, axis=0)
    energy_dist = make_dist('energy', particle_name, NP, density, energy_axis=Eaxis,
                            distvals=Edist, pitch_range=[-0.5,0.5])

    # Make tabulated speed distribution
    v_axis = np.linspace(0, 1e7, 500)
    v_dist = np.zeros_like(v_axis)
    v_dist[:250] = v_axis[:250]
    v_dist[250:] = v_axis[-1] - v_axis[250:]
    v_dist = np.repeat(np.atleast_2d(v_dist), NP, axis=0)
    speed_dist = make_dist('speed', particle_name, NP, density, speed_axis=v_axis,
                            distvals=v_dist, pitch_range=[-0.5,0.5])



import os

import numpy as np
import scipy.interpolate as interp
from scipy.special import legendre

from dress.config import cross_section_dir
from dress.reactions.particle import Particle
from dress.reactions.reaction import Reaction


# Load Legendre coefficients for DD differential cross section
c_endf = np.loadtxt(os.path.join(cross_section_dir,'ddn3he_legendre_endf.txt')).T

# Convert from LAB to COM (equal masses) and energies to keV
c_endf[0,:] = c_endf[0,:]/2000.0

# Number of coefficients in the expansion
n_endf = c_endf.shape[0] - 1


class DDNHe3Reaction(Reaction):
    """Class representing the d + d -> n + he3 fusion reaction."""

    __slots__ = ()

    def __init__(self):

        super().__init__('d', 'd', 'n', 'he3', None)

    
    def _calc_sigma_tot(self, E):
        E = np.atleast_1d(E).astype('d')
        sigma = np.zeros_like(E)
        
        # Bosch-Hale parameterization of the D(d,n)3He reaction
        B_G = 31.3970
            
        A1  = 5.3701e4
        A2  = 3.3027e2
        A3  = -1.2706e-1
        A4  = 2.9327e-5
        A5  = -2.5151e-9
        
        # Calculate the S factor
        S = A1+E*(A2+E*(A3+E*(A4+E*A5)))

        # Calculate the cross section
        nonzero = E > 0
        sigma[nonzero] = S[nonzero]/(E[nonzero]*np.exp(B_G/np.sqrt(E[nonzero])))     # mb
        
        return sigma * 1e-31     # m**2


    def _calc_sigma_diff(self, E, costheta):
        E = np.atleast_1d(E)
        costheta = np.atleast_1d(costheta)

        # Compute the angular dependence of the DD reaction
        A     = interp.interp1d(c_endf[0],c_endf[1:], fill_value=0, bounds_error=False)(E)
        prob  = 0.5*np.ones_like(E)
        
        for i in range(n_endf):
            l = i+1
            P = legendre(l)(costheta)
            prob += (2*l+1)/2 * A[i] * P

        sigma = prob*self.calc_sigma_tot(E)/(2*np.pi)    # m**2/sr

        return sigma



"""Tools for plotting spectra and distributions in the `dress` framework."""

import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import numpy as np
import scipy.constants as const

from dress import relkin, vec_ops

E_conversion_factors = {}
E_conversion_factors['keV'] = 1.0
E_conversion_factors['MeV'] = 1e-3


def convert_energy(E, unit):
    try:
        cf = E_conversion_factors[unit]
    except KeyError:
        raise ValueError(f'Conversion to {unit} currently not supported')

    return E*cf


def plot_spec(spec, *bin_edges, **kwargs):
    """Plot a given spectrum.

    Parameters
    ----------
    spec : spectrum array
        Spectrum produced by a SpectrumCalculator.

    bin_edges: arrays
        Energy bins and (optionally) pitch bins for the product particle.

    *** Possible keyword arguments: ***

    E_label : str
        Label for the energy axis. Default is "E".

    A_label : str
        Label for the pitch axis, if present. Default is "cos(θ)".

    spec_label : str
        Label for the spectrum axis. Default is "spectrum (events/bin/s)".

    label : str
        Label for the legend. Default is None (no legend will be drawn).

    E_unit : str
        Which energy units to use. Default is "MeV".
    
    convert_E : bool
        Whether to convert the energy axis according to the value of `E_unit`, under 
        the assumption that the input energy axis is in keV (the default units of dress). 
        Default is True.

    erase : bool
        Whether to erase existing spectra from the figure. Default is `True`.
    """

    # Get keyword arguments
    E_label = kwargs.get('E_label', 'E')
    A_label = kwargs.get('A_label', 'cos(θ)')
    spec_label = kwargs.get('spec_label', 'events/bin/s')
    label = kwargs.get('label', None)
    E_unit = kwargs.get('E_unit', 'MeV')
    convert_E = kwargs.get('convert_E', True)
    erase = kwargs.get('erase', True)

    # Convert energy scale, if necessary
    E_bins = bin_edges[0]
    if convert_E:
        E_bins = convert_energy(E_bins, E_unit)

    # Add energy unit to the energy label
    E_label += f' ({E_unit})'

    if len(bin_edges) == 1:
        plt.figure('DRESS energy spectrum')
        if erase: plt.clf()
        plt.step(E_bins[1:], spec, where='pre', label=label)
        plt.xlabel(E_label)
        plt.ylabel(spec_label)
        plt.legend()
        
    if len(bin_edges) == 2:
        plt.figure('DRESS energy-pitch spectrum')
        if erase: plt.clf()
        A_bins = bin_edges[1]
        plt.pcolor(E_bins, A_bins, spec.T)
        plt.xlabel(E_label)
        plt.ylabel(A_label)
        plt.colorbar(label=spec_label)
        
    
def plot_emissivity(vols, spec, *bin_edges, **kwargs):
    """Plot 2D spatial emissivity from given volume elements.

    Parameters
    ----------
    vols : list of instances of dress.volspec.VolumeElement
        The volume elements for which the spectra has been calculated.
    
    spec : spectrum array
        This should be a spectrum of the kind calculated with dress.volspec.calc_vols, 
        using the `integrate=False` option.

    bin_edges : arrays
        Energy bins and (optionally) pitch bins for the product particle.

    x_label, y_label : str
        Label for the spatial x and y- axes.

    ems_label : str
        Label for the emissivity axis. Default is "events/m³/s"."""

    if len(vols.pos) != 2:
        raise ValueError('Emissivity plot currently only works with 2D profiles')

    # Keyword arguments
    x_label = kwargs.get('x_label', 'x')
    y_label = kwargs.get('y_label', 'y')
    ems_label = kwargs.get('ems_label', 'events/m³/s')

    # Extract data to plot
    x = np.atleast_1d(vols.pos[0])
    y = np.atleast_1d(vols.pos[1])

    if spec.ndim == 2:
        sum_axis = 1
    elif spec.ndim == 3:
        sum_axis = (1,2)

    ems = spec.sum(axis=sum_axis)

    plt.figure('DRESS emissivity')
    plt.clf()
    plt.tripcolor(x, y, ems)
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.colorbar(label=ems_label)
    plt.axis('equal')
    

def plot_dist_point(dist, i_spatial=1, n_samples=100_000, dist_type='energy-pitch', 
                    log_dist= False, bins=50):
    """Plot distribution at a given spatial location.

    The plotting is done by sampling `n_samples` velocities and making a histogram,
    thus generating a plot that exactly represent the distribution as 'seen' by
    the spectrum calculation routines in `dress`.

    Parameters
    ----------
    dist : dress.dists.VelocityDistribution
        The distribution to plot.

    i_spatial : int
        The spatial point for which the distribution is to be plotted.

    n_samples : int
        Number of samples to draw for the histogram.

    dist_type : str
        How to visualize the distribution. Can be one of 
            - `energy-pitch`
            - `vpar-vperp`
            - `energy`
            - `speed`

    log_dist : bool
        Whether to plot dist in log scale (default is False).

    bins : various
        Bins specification as expected by np.histogram and np.histogram2d."""

    d = _extract_dist(dist, i_spatial=i_spatial, n_samples=n_samples, 
                      dist_type=dist_type, bins=bins)

    # Plot settings
    if log_dist:
        norm = LogNorm()
    else:
        norm = None

    if dist_type == 'energy-pitch':
        plt.figure('energy-pitch dist')
        plt.clf()
        plt.pcolor(d[1], d[2], d[0].T, norm=norm)
        plt.ylim(-1,1)
        plt.xlabel('Energy (keV)')
        plt.ylabel('v$_{\parallel}$/v (m/s)')
        plt.colorbar()

    elif dist_type == 'energy':
        plt.figure('energy dist')
        plt.clf()
        x = 0.5*(d[1][1:] + d[1][:-1])
        w = np.diff(d[1])
        plt.bar(x, d[0], width=w, log=log_dist)
        plt.xlabel('Energy (keV)')

    elif dist_type == 'vpar-vperp':
        plt.figure('vpar-vperp dist')
        plt.clf()
        plt.pcolor(d[1], d[2], d[0].T, norm=norm)
        plt.xlabel('v$_{\parallel}$ (m/s)')
        plt.ylabel('v$_{\perp}$ (m/s)')
        plt.ylim(bottom=0)
        plt.axis('equal')
        plt.colorbar()

    elif dist_type == 'speed':
        plt.figure('speed dist')
        plt.clf()
        x = 0.5*(d[1][1:] + d[1][:-1])
        w = np.diff(d[1])
        plt.bar(x, d[0], width=w, log=log_dist)
        plt.xlabel('Speed (m/s)')


def _extract_dist(dist, i_spatial=1, n_samples=100_000, dist_type='energy-pitch', bins=50):
    """Sample distribution at the given spatial point 
    and generate histogram in the requested variables.

    Possible choices for `dist_type` are
        - `energy-pitch`
        - `vpar-vperp`
        - `energy`
        - `speed`"""

    # Sample velocities from the distribution
    vel = dist.sample(n_samples, index=i_spatial)
    density = dist.density[i_spatial]

    # Convert to the appropriate coordinates
    m = dist.particle.m     # mass i keV/c**2

    if dist_type == 'energy-pitch':
        E = relkin.get_energy(vel, m)
        ref_dir = dist.ref_dir[:,i_spatial]
        v_par = np.sum(vel*ref_dir[:,None], axis=0)
        v = np.sqrt(vec_ops.dot(vel, vel))
        pitch = v_par / v
        h = np.histogram2d(E, pitch, density=True, bins=bins)
        
    elif dist_type == 'energy':
        E = relkin.get_energy(vel, m)
        h = np.histogram(E, density=True, bins=bins)
        
    elif dist_type == 'vpar-vperp':
        ref_dir = dist.ref_dir[:,i_spatial]
        v_par = np.sum(vel*ref_dir[:,None], axis=0)
        vel_par = ref_dir[:,None] * v_par[None,:]
        vel_perp = vel - vel_par
        v_perp = np.sqrt(vec_ops.dot(vel_perp, vel_perp))
        h = np.histogram2d(v_par, v_perp, density=True, bins=bins)

    elif dist_type == 'speed':
        v = np.sqrt(vec_ops.dot(vel, vel))
        h = np.histogram(v, density=True, bins=bins)

    d = h[0] * density
    
    return (d, *h[1:])

        
def explore_dist(dist, **kwargs):
    """Interactive distribution plots.

    The particle density as a function of posistion is plotted.
    By clocking in this figure the user can plot the velocity distribution
    at the different points.

    Keyword arguments are passed to plot_dist_point."""

    if dist.pos is None:
        raise ValueError('Must set pos attribute')

    if len(dist.pos) > 2:
        print('Only 2D plotting supported so far. Plotting density projected on first two dimensions.')

    # Plot density
    density_fig = plt.figure('density')
    plt.tripcolor(dist.pos[0], dist.pos[1], dist.density)
    plt.axis('equal')

    click_fun = lambda event: _density_click_fun(event, dist, **kwargs)
    cid = density_fig.canvas.mpl_connect('button_press_event', click_fun)    # connect to event manager

    return density_fig, click_fun
   
 
def _density_click_fun(event, dist, **kwargs):
    """Determine what happens after clicking in the density plot."""

    # Find the spatial point closest to the click
    x_click = event.xdata
    y_click = event.ydata

    distance = np.sqrt( (x_click - dist.pos[0])**2 + (y_click - dist.pos[1])**2 )

    i_click = np.argmin(distance)

    plot_dist_point(dist, i_spatial=i_click, **kwargs)
    plt.draw()
